"""Tests the training loop."""

import os
import shutil

import pandas as pd
from datasets import load_dataset
from sentence_transformers import (
    SentenceTransformerTrainer,
    SentenceTransformerTrainingArguments,
)
from sentence_transformers.training_args import BatchSamplers

from pylate import evaluation, losses, models, utils


def test_contrastive_training() -> None:
    """Test constrastive training."""
    if os.path.exists(path="tests/contrastive"):
        shutil.rmtree("tests/contrastive")

    model = models.ColBERT(model_name_or_path="sentence-transformers/all-MiniLM-L6-v2")

    dataset = load_dataset("lightonai/lighton-ms-marco-mini", "triplet", split="train")

    splits = dataset.train_test_split(test_size=0.5)

    train_dataset, eval_dataset = splits["train"], splits["test"]

    train_loss = losses.Contrastive(model=model)

    dev_evaluation = evaluation.ColBERTTripletEvaluator(
        anchors=eval_dataset["query"],
        positives=eval_dataset["positive"],
        negatives=eval_dataset["negative"],
    )

    args = SentenceTransformerTrainingArguments(
        output_dir="tests/contrastive",
        overwrite_output_dir=True,
        num_train_epochs=1,
        per_device_train_batch_size=1,
        per_device_eval_batch_size=1,
        fp16=False,
        bf16=False,
        batch_sampler=BatchSamplers.NO_DUPLICATES,
        eval_strategy="steps",
        eval_steps=1,
        save_strategy="epoch",
        save_steps=1,
        save_total_limit=1,
        learning_rate=3e-6,
        do_eval=True,
    )

    trainer = SentenceTransformerTrainer(
        model=model,
        args=args,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        loss=train_loss,
        evaluator=dev_evaluation,
        data_collator=utils.ColBERTCollator(tokenize_fn=model.tokenize),
    )

    trainer.train()

    assert os.path.isdir("tests/contrastive")

    metrics = dev_evaluation(
        model=model,
        output_path="tests/contrastive/",
    )

    assert isinstance(metrics, dict)

    assert os.path.isfile(path="tests/contrastive/triplet_evaluation_results.csv")

    results = pd.read_csv(
        filepath_or_buffer="tests/contrastive/triplet_evaluation_results.csv"
    )

    assert "accuracy" in list(results.columns)

    if os.path.exists(path="tests/contrastive"):
        shutil.rmtree("tests/contrastive")



import itertools
import os

import numpy as np
import torch
from sqlitedict import SqliteDict
from voyager import Index, Space

from ..utils import iter_batch
from .base import Base


def reshape_embeddings(
    embeddings: np.ndarray | torch.Tensor,
) -> np.ndarray | torch.Tensor:
    """Reshape the embeddings, the Voyager index expects arrays with shape batch_size, n_tokens, embedding_size."""
    if isinstance(embeddings, np.ndarray):
        if len(embeddings.shape) == 2:
            return np.expand_dims(a=embeddings, axis=0)

    if isinstance(embeddings, torch.Tensor):
        return reshape_embeddings(embeddings=embeddings.cpu().detach().numpy())

    if isinstance(embeddings, list) and isinstance(embeddings[0], torch.Tensor):
        return [embedding.cpu().detach().numpy() for embedding in embeddings]

    return embeddings


class Voyager(Base):
    """Voyager index. The Voyager index is a fast and efficient index for approximate nearest neighbor search.

    Parameters
    ----------
    name
        The name of the collection.
    override
        Whether to override the collection if it already exists.
    embedding_size
        The number of dimensions of the embeddings.
    M
        The number of subquantizers.
    ef_construction
        The number of candidates to evaluate during the construction of the index.
    ef_search
        The number of candidates to evaluate during the search.

    Examples
    --------
    >>> from pylate import indexes, models

    >>> index = indexes.Voyager(
    ...     index_folder="test_indexes",
    ...     index_name="colbert",
    ...     override=True,
    ...     embedding_size=128,
    ... )

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2",
    ... )

    >>> documents_embeddings = model.encode(
    ...     ["fruits are healthy.", "fruits are good for health.", "fruits are bad for health."],
    ...     is_query=False,
    ... )

    >>> index = index.add_documents(
    ...     documents_ids=["1", "2", "3"],
    ...     documents_embeddings=documents_embeddings
    ... )

    >>> queries_embeddings = model.encode(
    ...     ["fruits are healthy.", "fruits are good for health and fun."],
    ...     is_query=True,
    ... )

    >>> matchs = index(queries_embeddings, k=30)

    >>> assert matchs["distances"].shape[0] == 2
    >>> assert isinstance(matchs, dict)
    >>> assert "documents_ids" in matchs
    >>> assert "distances" in matchs

    >>> queries_embeddings = model.encode(
    ...     "fruits are healthy.",
    ...     is_query=True,
    ... )

    >>> matchs = index(queries_embeddings, k=30)

    >>> assert matchs["distances"].shape[0] == 1
    >>> assert isinstance(matchs, dict)
    >>> assert "documents_ids" in matchs
    >>> assert "distances" in matchs

    """

    def __init__(
        self,
        index_folder: str = "indexes",
        index_name: str = "colbert",
        override: bool = False,
        embedding_size: int = 128,
        M: int = 64,
        ef_construction: int = 200,
        ef_search: int = 200,
    ) -> None:
        self.ef_search = ef_search

        if not os.path.exists(path=index_folder):
            os.makedirs(name=index_folder)

        self.index_path = os.path.join(index_folder, f"{index_name}.voyager")
        self.documents_ids_to_embeddings_path = os.path.join(
            index_folder, f"{index_name}_document_ids_to_embeddings.sqlite"
        )
        self.embeddings_to_documents_ids_path = os.path.join(
            index_folder, f"{index_name}_embeddings_to_documents_ids.sqlite"
        )

        self.index = self._create_collection(
            index_path=self.index_path,
            embedding_size=embedding_size,
            M=M,
            ef_constructions=ef_construction,
            override=override,
        )

    def _load_documents_ids_to_embeddings(self) -> SqliteDict:
        """Load the SQLite database that maps document IDs to embeddings."""
        return SqliteDict(self.documents_ids_to_embeddings_path, outer_stack=False)

    def _load_embeddings_to_documents_ids(self) -> SqliteDict:
        """Load the SQLite database that maps embeddings to document IDs."""
        return SqliteDict(self.embeddings_to_documents_ids_path, outer_stack=False)

    def _create_collection(
        self,
        index_path: str,
        embedding_size: int,
        M: int,
        ef_constructions: int,
        override: bool,
    ) -> None:
        """Create a new Voyager collection.

        Parameters
        ----------
        index_path
            The path to the index.
        embedding_size
            The size of the embeddings.
        M
            The number of subquantizers.
        ef_constructions
            The number of candidates to evaluate during the construction of the index.
        override
            Whether to override the collection if it already exists.

        """
        if os.path.exists(path=index_path) and not override:
            return Index.load(index_path)

        if os.path.exists(path=index_path):
            os.remove(index_path)

        # Create the Voyager index
        index = Index(
            Space.Cosine,
            num_dimensions=embedding_size,
            M=M,
            ef_construction=ef_constructions,
        )

        index.save(index_path)

        if override and os.path.exists(path=self.documents_ids_to_embeddings_path):
            os.remove(path=self.documents_ids_to_embeddings_path)

        if override and os.path.exists(path=self.embeddings_to_documents_ids_path):
            os.remove(path=self.embeddings_to_documents_ids_path)

        # Create the SQLite databases
        documents_ids_to_embeddings = self._load_documents_ids_to_embeddings()
        documents_ids_to_embeddings.close()

        embeddings_to_documents_ids = self._load_embeddings_to_documents_ids()
        embeddings_to_documents_ids.close()

        return index

    def add_documents(
        self,
        documents_ids: str | list[str],
        documents_embeddings: list[np.ndarray | torch.Tensor],
        batch_size: int = 2000,
    ) -> None:
        """Add documents to the index."""
        if isinstance(documents_ids, str):
            documents_ids = [documents_ids]

        """Add documents to the index."""
        documents_embeddings = reshape_embeddings(embeddings=documents_embeddings)

        documents_ids_to_embeddings = self._load_documents_ids_to_embeddings()
        embeddings_to_documents_ids = self._load_embeddings_to_documents_ids()
        for document_embeddings_batch, documents_ids_batch in zip(
            iter_batch(
                documents_embeddings,
                batch_size,
                desc=f"Adding documents to the index (bs={batch_size})",
            ),
            iter_batch(documents_ids, batch_size, tqdm_bar=False),
        ):
            embeddings_ids = self.index.add_items(
                list(itertools.chain(*document_embeddings_batch))
            )

            total = 0
            for doc_id, document_embeddings in zip(
                documents_ids_batch, document_embeddings_batch
            ):
                document_embeddings_ids = embeddings_ids[
                    total : total + len(document_embeddings)
                ]
                documents_ids_to_embeddings[doc_id] = document_embeddings_ids

                embeddings_to_documents_ids.update(
                    dict.fromkeys(document_embeddings_ids, doc_id)
                )
                total += len(document_embeddings)

        documents_ids_to_embeddings.commit()
        documents_ids_to_embeddings.close()

        embeddings_to_documents_ids.commit()
        embeddings_to_documents_ids.close()
        self.index.save(self.index_path)
        return self

    def remove_documents(self, documents_ids: list[str]) -> None:
        """Remove documents from the index.

        Parameters
        ----------
        documents_ids
            The documents IDs to remove.

        """
        documents_ids_to_embeddings = self._load_documents_ids_to_embeddings()
        embeddings_to_documents_ids = self._load_embeddings_to_documents_ids()

        for document_id in documents_ids:
            embeddings_ids = documents_ids_to_embeddings[document_id]
            for embedding_id in embeddings_ids:
                del embeddings_to_documents_ids[embedding_id]
                self.index.mark_deleted(embedding_id)
            del documents_ids_to_embeddings[document_id]

        documents_ids_to_embeddings.commit()
        embeddings_to_documents_ids.commit()

        documents_ids_to_embeddings.close()
        embeddings_to_documents_ids.close()
        self.index.save(self.index_path)
        return self

    def __call__(
        self,
        queries_embeddings: np.ndarray | torch.Tensor,
        k: int = 10,
    ) -> dict:
        """Query the index for the nearest neighbors of the queries embeddings.

        Parameters
        ----------
        queries_embeddings
            The queries embeddings.
        k
            The number of nearest neighbors to return.

        """
        embeddings_to_documents_ids = self._load_embeddings_to_documents_ids()
        k = min(k, len(embeddings_to_documents_ids))

        queries_embeddings = reshape_embeddings(embeddings=queries_embeddings)
        n_queries = len(queries_embeddings)

        indices, distances = self.index.query(
            list(itertools.chain(*queries_embeddings)), k, query_ef=self.ef_search
        )

        if len(indices) == 0:
            raise ValueError("Index is empty, add documents before querying.")

        documents = [
            [
                [
                    embeddings_to_documents_ids[str(token_indice)]
                    for token_indice in tokens_indices
                ]
                for tokens_indices in document_indices
            ]
            for document_indices in indices.reshape(n_queries, -1, k)
        ]

        embeddings_to_documents_ids.close()

        return {
            "documents_ids": documents,
            "distances": distances.reshape(n_queries, -1, k),
        }

    def get_documents_embeddings(
        self, document_ids: list[list[str]]
    ) -> list[list[list[int | float]]]:
        """Retrieve document embeddings for re-ranking from Voyager."""

        # Load mappings of document IDs to embedding IDs
        documents_ids_to_embeddings = self._load_documents_ids_to_embeddings()

        # Retrieve embedding IDs in the same structure as document IDs
        embedding_ids_structure = [
            [documents_ids_to_embeddings[doc_id] for doc_id in doc_group]
            for doc_group in document_ids
        ]

        documents_ids_to_embeddings.close()

        # Flatten the embedding IDs for a single API call
        flattened_embedding_ids = list(
            itertools.chain.from_iterable(
                itertools.chain.from_iterable(embedding_ids_structure)
            )
        )

        # Retrieve all embeddings in one API call
        all_embeddings = self.index.get_vectors(flattened_embedding_ids)

        # Reconstruct embeddings into the original structure
        reconstructed_embeddings = []
        embedding_idx = 0
        for group_embedding_ids in embedding_ids_structure:
            group_embeddings = []
            for doc_embedding_ids in group_embedding_ids:
                num_embeddings = len(doc_embedding_ids)
                group_embeddings.append(
                    all_embeddings[embedding_idx : embedding_idx + num_embeddings]
                )
                embedding_idx += num_embeddings
            reconstructed_embeddings.append(group_embeddings)

        return reconstructed_embeddings



from typing import Callable, Iterable

import torch

from ..models import ColBERT
from ..scores import colbert_kd_scores
from .contrastive import extract_skiplist_mask


class Distillation(torch.nn.Module):
    """Distillation loss for ColBERT model. The loss is computed with respect to the format of SentenceTransformer library.

    Parameters
    ----------
    model
        SentenceTransformer model.
    score_metric
        Function that returns a score between two sequences of embeddings.
    size_average
        Average by the size of the mini-batch or perform sum.

    Examples
    --------
    >>> from pylate import models, losses

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> distillation = losses.Distillation(model=model)

    >>> query = model.tokenize([
    ...     "fruits are healthy.",
    ... ], is_query=True)

    >>> documents = model.tokenize([
    ...     "fruits are good for health.",
    ...     "fruits are bad for health."
    ... ], is_query=False)

    >>> sentence_features = [query, documents]

    >>> labels = torch.tensor([
    ...     [0.7, 0.3],
    ... ], dtype=torch.float32)

    >>> loss = distillation(sentence_features=sentence_features, labels=labels)

    >>> assert isinstance(loss.item(), float)
    """

    def __init__(
        self,
        model: ColBERT,
        score_metric: Callable = colbert_kd_scores,
        size_average: bool = True,
        normalize_scores: bool = True,
    ) -> None:
        super(Distillation, self).__init__()
        self.score_metric = score_metric
        self.model = model
        self.loss_function = torch.nn.KLDivLoss(
            reduction="batchmean" if size_average else "sum", log_target=True
        )
        self.normalize_scores = normalize_scores

    def forward(
        self, sentence_features: Iterable[dict[str, torch.Tensor]], labels: torch.Tensor
    ) -> torch.Tensor:
        """Computes the distillation loss with respect to SentenceTransformer.

        Parameters
        ----------
        sentence_features
            List of tokenized sentences. The first sentence is the query and the rest are documents.
        labels
            The logits for the distillation loss.

        """
        queries_embeddings = torch.nn.functional.normalize(
            self.model(sentence_features[0])["token_embeddings"], p=2, dim=-1
        )
        # Compute the bs * n_ways embeddings
        documents_embeddings = torch.nn.functional.normalize(
            self.model(sentence_features[1])["token_embeddings"], p=2, dim=-1
        )

        # Reshape them to (bs, n_ways)
        documents_embeddings = documents_embeddings.view(
            queries_embeddings.size(0), -1, *documents_embeddings.shape[1:]
        )

        # handle the model being wrapped in (D)DP and so require to access module first
        skiplist = (
            self.model.skiplist
            if hasattr(self.model, "skiplist")
            else self.model.module.skiplist
        )

        masks = extract_skiplist_mask(
            sentence_features=sentence_features, skiplist=skiplist
        )

        documents_embeddings_mask = masks[1].view(
            queries_embeddings.size(0), -1, *masks[1].shape[1:]
        )
        scores = self.score_metric(
            queries_embeddings,
            documents_embeddings,
            documents_embeddings_mask,
        )
        if self.normalize_scores:
            # Compute max and min along the num_scores dimension (dim=1)
            max_scores, _ = torch.max(scores, dim=1, keepdim=True)
            min_scores, _ = torch.min(scores, dim=1, keepdim=True)

            # Avoid division by zero by adding a small epsilon
            epsilon = 1e-8

            # Normalize the scores
            scores = (scores - min_scores) / (max_scores - min_scores + epsilon)
        return self.loss_function(
            torch.nn.functional.log_softmax(scores, dim=-1),
            torch.nn.functional.log_softmax(labels, dim=-1),
        )



import numpy as np
import torch

from ..scores import colbert_scores
from ..utils import convert_to_tensor as func_convert_to_tensor


def reshape_embeddings(
    embeddings: np.ndarray | torch.Tensor,
) -> np.ndarray | torch.Tensor:
    """Reshape the embeddings to the correct shape."""
    if isinstance(embeddings, torch.Tensor):
        if embeddings.ndim == 2:
            embeddings = embeddings.unsqueeze(dim=0)

    elif isinstance(embeddings, np.ndarray):
        if len(embeddings.shape) == 2:
            return np.expand_dims(a=embeddings, axis=0)

    return embeddings


def rerank(
    documents_ids: list[list[int | str]],
    queries_embeddings: list[list[float | int] | np.ndarray | torch.Tensor],
    documents_embeddings: list[list[float | int] | np.ndarray | torch.Tensor],
    device: str = None,
) -> list[list[dict[str, float]]]:
    """Rerank the documents based on the queries embeddings.

    Parameters
    ----------
    queries
        The queries.
    documents_ids
        The documents ids.
    queries_embeddings
        The queries embeddings which is a dictionary of queries and their embeddings.
    documents_embeddings
        The documents embeddings which is a dictionary of documents ids and their embeddings.

    Examples
    --------
    >>> from pylate import models, rank

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> queries = [
    ...     "query A",
    ...     "query B",
    ... ]

    >>> documents = [
    ...     ["document A", "document B"],
    ...     ["document 1", "document C", "document B"],
    ... ]

    >>> documents_ids = [
    ...    [1, 2],
    ...    [1, 3, 2],
    ... ]

    >>> queries_embeddings = model.encode(
    ...     queries,
    ...     is_query=True,
    ...     batch_size=1,
    ... )

    >>> documents_embeddings = model.encode(
    ...     documents,
    ...     is_query=False,
    ...     batch_size=1,
    ... )

    >>> reranked_documents = rank.rerank(
    ...     documents_ids=documents_ids,
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> assert isinstance(reranked_documents, list)
    >>> assert len(reranked_documents) == 2
    >>> assert len(reranked_documents[0]) == 2
    >>> assert len(reranked_documents[1]) == 3
    >>> assert isinstance(reranked_documents[0], list)
    >>> assert isinstance(reranked_documents[0][0], dict)
    >>> assert "id" in reranked_documents[0][0]
    >>> assert "score" in reranked_documents[0][0]

    """
    results = []

    queries_embeddings = reshape_embeddings(embeddings=queries_embeddings)
    documents_embeddings = reshape_embeddings(embeddings=documents_embeddings)

    for query_embeddings, query_documents_ids, query_documents_embeddings in zip(
        queries_embeddings, documents_ids, documents_embeddings
    ):
        query_embeddings = func_convert_to_tensor(query_embeddings)

        query_documents_embeddings = [
            func_convert_to_tensor(query_document_embeddings)
            for query_document_embeddings in query_documents_embeddings
        ]

        # Pad the documents embeddings
        query_documents_embeddings = torch.nn.utils.rnn.pad_sequence(
            query_documents_embeddings, batch_first=True, padding_value=0
        )

        if device is not None:
            query_embeddings = query_embeddings.to(device)
            query_documents_embeddings = query_documents_embeddings.to(device)
        else:
            query_documents_embeddings = query_documents_embeddings.to(
                query_embeddings.device
            )

        query_scores = colbert_scores(
            queries_embeddings=query_embeddings.unsqueeze(0),
            documents_embeddings=query_documents_embeddings,
        )[0]

        scores, sorted_indices = torch.sort(input=query_scores, descending=True)
        scores = scores.cpu().tolist()

        query_documents = [query_documents_ids[idx] for idx in sorted_indices.tolist()]

        results.append(
            [
                {"id": doc_id, "score": score}
                for doc_id, score in zip(query_documents, scores)
            ]
        )

    return results



import numpy as np
import torch

from ..utils.tensor import convert_to_tensor


def colbert_scores(
    queries_embeddings: list | np.ndarray | torch.Tensor,
    documents_embeddings: list | np.ndarray | torch.Tensor,
    mask: torch.Tensor = None,
) -> torch.Tensor:
    """Computes the ColBERT scores between queries and documents embeddings. The score is computed as the sum of maximum similarities
    between the query and the document.

    Parameters
    ----------
    queries_embeddings
        The first tensor. The queries embeddings. Shape: (batch_size, num tokens queries, embedding_size)
    documents_embeddings
        The second tensor. The documents embeddings. Shape: (batch_size, num tokens documents, embedding_size)

    Examples
    --------
    >>> import torch

    >>> queries_embeddings = torch.tensor([
    ...     [[1.], [0.], [0.], [0.]],
    ...     [[0.], [2.], [0.], [0.]],
    ...     [[0.], [0.], [3.], [0.]],
    ... ])

    >>> documents_embeddings = torch.tensor([
    ...     [[10.], [0.], [1.]],
    ...     [[0.], [100.], [1.]],
    ...     [[1.], [0.], [1000.]],
    ... ])

    >>> scores = colbert_scores(
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings
    ... )

    >>> scores
    tensor([[  10.,  100., 1000.],
            [  20.,  200., 2000.],
            [  30.,  300., 3000.]])

    """
    queries_embeddings = convert_to_tensor(queries_embeddings)
    documents_embeddings = convert_to_tensor(documents_embeddings)

    scores = torch.einsum(
        "ash,bth->abst",
        queries_embeddings,
        documents_embeddings,
    )

    if mask is not None:
        mask = convert_to_tensor(mask)
        scores = scores * mask.unsqueeze(0).unsqueeze(2)

    return scores.max(axis=-1).values.sum(axis=-1)


def colbert_scores_pairwise(
    queries_embeddings: torch.Tensor,
    documents_embeddings: torch.Tensor,
) -> torch.Tensor:
    """Computes the ColBERT score for each query-document pair. The score is computed as the sum of maximum similarities
    between the query and the document for corresponding pairs.

    Parameters
    ----------
    queries_embeddings
        The first tensor. The queries embeddings. Shape: (batch_size, num tokens queries, embedding_size)
    documents_embeddings
        The second tensor. The documents embeddings. Shape: (batch_size, num tokens documents, embedding_size)

    Examples
    --------
    >>> import torch

    >>> queries_embeddings = torch.tensor([
    ...     [[1.], [0.], [0.], [0.]],
    ...     [[0.], [2.], [0.], [0.]],
    ...     [[0.], [0.], [3.], [0.]],
    ... ])

    >>> documents_embeddings = torch.tensor([
    ...     [[10.], [0.], [1.]],
    ...     [[0.], [100.], [1.]],
    ...     [[1.], [0.], [1000.]],
    ... ])

    >>> scores = colbert_scores_pairwise(
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings
    ... )

    >>> scores
    tensor([  10.,  200., 3000.])

    """
    scores = []

    for query_embedding, document_embedding in zip(
        queries_embeddings, documents_embeddings
    ):
        query_embedding = convert_to_tensor(query_embedding)
        document_embedding = convert_to_tensor(document_embedding)

        query_document_score = torch.einsum(
            "sh,th->st",
            query_embedding,
            document_embedding,
        )

        scores.append(query_document_score.max(axis=-1).values.sum())

    return torch.stack(scores, dim=0)


def colbert_kd_scores(
    queries_embeddings: list | np.ndarray | torch.Tensor,
    documents_embeddings: list | np.ndarray | torch.Tensor,
    mask: torch.Tensor = None,
) -> torch.Tensor:
    """Computes the ColBERT scores between queries and documents embeddings. This scoring function is dedicated to the knowledge distillation pipeline.

    Examples
    --------
    >>> import torch

    >>> queries_embeddings = torch.tensor([
    ...     [[1.], [0.], [0.], [0.]],
    ...     [[0.], [2.], [0.], [0.]],
    ...     [[0.], [0.], [3.], [0.]],
    ... ])

    >>> documents_embeddings = torch.tensor([
    ...     [[[10.], [0.], [1.]], [[20.], [0.], [1.]], [[30.], [0.], [1.]]],
    ...     [[[0.], [100.], [1.]], [[0.], [200.], [1.]], [[0.], [300.], [1.]]],
    ...     [[[1.], [0.], [1000.]], [[1.], [0.], [2000.]], [[10.], [0.], [3000.]]],
    ... ])
    >>> mask = torch.tensor([
    ...     [[1., 1., 1.], [1., 1., 1.], [1., 1., 1.]],
    ...     [[1., 1., 1.], [1., 1., 1.], [1., 1., 1.]],
    ...     [[1., 1., 1.], [1., 1., 1.], [1., 1., 0.]],
    ... ])
    >>> colbert_kd_scores(
    ...     queries_embeddings=queries_embeddings,
    ...     documents_embeddings=documents_embeddings,
    ...     mask=mask
    ... )
    tensor([[  10.,   20.,   30.],
            [ 200.,  400.,  600.],
            [3000., 6000., 30.]])

    """
    queries_embeddings = convert_to_tensor(queries_embeddings)
    documents_embeddings = convert_to_tensor(documents_embeddings)

    scores = torch.einsum(
        "ash,abth->abst",
        queries_embeddings,
        documents_embeddings,
    )
    if mask is not None:
        mask = convert_to_tensor(mask)
        scores = scores * mask.unsqueeze(2)

    return scores.max(axis=-1).values.sum(axis=-1)



"""Relevant constants for the project."""

# List of models that do not have a family in the Hugging Face model hub.
HUGGINGFACE_MODELS = [
    "albert-base-v1",
    "albert-base-v2",
    "albert-large-v1",
    "albert-large-v2",
    "albert-xlarge-v1",
    "albert-xlarge-v2",
    "albert-xxlarge-v1",
    "albert-xxlarge-v2",
    "bert-base-cased-finetuned-mrpc",
    "bert-base-cased",
    "bert-base-chinese",
    "bert-base-german-cased",
    "bert-base-german-dbmdz-cased",
    "bert-base-german-dbmdz-uncased",
    "bert-base-multilingual-cased",
    "bert-base-multilingual-uncased",
    "bert-base-uncased",
    "bert-large-cased-whole-word-masking-finetuned-squad",
    "bert-large-cased-whole-word-masking",
    "bert-large-cased",
    "bert-large-uncased-whole-word-masking-finetuned-squad",
    "bert-large-uncased-whole-word-masking",
    "bert-large-uncased",
    "camembert-base",
    "ctrl",
    "distilbert-base-cased-distilled-squad",
    "distilbert-base-cased",
    "distilbert-base-german-cased",
    "distilbert-base-multilingual-cased",
    "distilbert-base-uncased-distilled-squad",
    "distilbert-base-uncased-finetuned-sst-2-english",
    "distilbert-base-uncased",
    "distilgpt2",
    "distilroberta-base",
    "gpt2-large",
    "gpt2-medium",
    "gpt2-xl",
    "gpt2",
    "openai-gpt",
    "roberta-base-openai-detector",
    "roberta-base",
    "roberta-large-mnli",
    "roberta-large-openai-detector",
    "roberta-large",
    "t5-11b",
    "t5-3b",
    "t5-base",
    "t5-large",
    "t5-small",
    "transfo-xl-wt103",
    "xlm-clm-ende-1024",
    "xlm-clm-enfr-1024",
    "xlm-mlm-100-1280",
    "xlm-mlm-17-1280",
    "xlm-mlm-en-2048",
    "xlm-mlm-ende-1024",
    "xlm-mlm-enfr-1024",
    "xlm-mlm-enro-1024",
    "xlm-mlm-tlm-xnli15-1024",
    "xlm-mlm-xnli15-1024",
    "xlm-roberta-base",
    "xlm-roberta-large-finetuned-conll02-dutch",
    "xlm-roberta-large-finetuned-conll02-spanish",
    "xlm-roberta-large-finetuned-conll03-english",
    "xlm-roberta-large-finetuned-conll03-german",
    "xlm-roberta-large",
    "xlnet-base-cased",
    "xlnet-large-cased",
]



from .beir import evaluate, get_beir_triples, load_beir
from .colbert_distillation import ColBERTDistillationEvaluator
from .colbert_triplet import ColBERTTripletEvaluator
from .custom_dataset import load_custom_dataset

__all__ = [
    "ColBERTTripletEvaluator",
    "ColBERTDistillationEvaluator",
    "get_beir_triples",
    "load_beir",
    "load_custom_dataset",
    "evaluate",
]



import tqdm


def iter_batch(
    X: list[str], batch_size: int, tqdm_bar: bool = True, desc: str = ""
) -> list:
    """Iterate over a list of elements by batch.

    Examples
    --------
    >>> from pylate import utils

    >>> X = [
    ...  "element 0",
    ...  "element 1",
    ...  "element 2",
    ...  "element 3",
    ...  "element 4",
    ... ]

    >>> n_samples = 0
    >>> for batch in utils.iter_batch(X, batch_size=2):
    ...     n_samples += len(batch)

    >>> n_samples
    5

    """
    batchs = [X[pos : pos + batch_size] for pos in range(0, len(X), batch_size)]

    if tqdm_bar:
        for batch in tqdm.tqdm(
            iterable=batchs,
            position=0,
            total=1 + len(X) // batch_size,
            desc=desc,
        ):
            yield batch
    else:
        yield from batchs



import itertools
import logging
import os
from contextlib import nullcontext

import torch
from sentence_transformers.evaluation.SentenceEvaluator import SentenceEvaluator
from sentence_transformers.SentenceTransformer import SentenceTransformer

from ..scores import colbert_kd_scores
from .colbert_triplet import csv_writer, evaluation_message

logger = logging.getLogger(__name__)


class ColBERTDistillationEvaluator(SentenceEvaluator):
    """ColBERT Distillation Evaluator.
    This class is used to monitor the distillation process of a ColBERT model.

    Parameters
    ----------
    queries
        Set of queries.
    documents
        Set of documents. Each query has a list of documents. Each document is a list of strings.
        Number of documents should be the same for each query.
    scores
        The scores associated with the documents. Each query / documents pairs has a list of scores.
    name
        The name of the evaluator.
    batch_size
        The batch size.
    show_progress_bar
        Whether to show the progress bar.
    write_csv
        Whether to write the results to a CSV file.
    truncate_dim
        The dimension to truncate the embeddings.


    Examples
    --------

    >>> from pylate import models, evaluation

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> queries = [
    ...     "query A",
    ...     "query B",
    ... ]

    >>> documents = [
    ...     ["document A", "document B", "document C"],
    ...     ["document C C", "document B B", "document A A"],
    ... ]

    >>> scores = [
    ...     [0.9, 0.1, 0.05],
    ...     [0.05, 0.9, 0.1],
    ... ]

    >>> distillation_evaluator = evaluation.ColBERTDistillationEvaluator(
    ...     queries=queries,
    ...     documents=documents,
    ...     scores=scores,
    ...     write_csv=True,
    ... )

    >>> results = distillation_evaluator(model=model, output_path=".")


    >>> assert "kl_divergence" in results
    >>> assert isinstance(results["kl_divergence"], float)

    >>> import pandas as pd
    >>> df = pd.read_csv(distillation_evaluator.csv_file)
    >>> assert df.columns.tolist() == distillation_evaluator.csv_headers

    """

    def __init__(
        self,
        queries: list[str],
        documents: list[list[str]],
        scores: list[list[float]],
        name: str = "",
        batch_size: int = 16,
        show_progress_bar: bool = False,
        write_csv: bool = True,
        truncate_dim: int | None = None,
        normalize_scores: bool = True,
    ) -> None:
        super().__init__()
        assert len(queries) == len(documents)
        self.queries = queries
        # Flatten the documents list
        self.documents = list(itertools.chain.from_iterable(documents))
        self.scores = scores
        self.name = name
        self.truncate_dim = truncate_dim
        self.batch_size = batch_size
        self.show_progress_bar = show_progress_bar
        self.write_csv = write_csv

        self.loss = torch.nn.KLDivLoss(reduction="batchmean", log_target=True)
        self.normalize_scores = normalize_scores
        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO
                or logger.getEffectiveLevel() == logging.DEBUG
            )

        self.csv_file: str = (
            "distillation_evaluation" + ("_" + name if name else "") + "_results.csv"
        )

        self.csv_headers = [
            "epoch",
            "steps",
            "kl_divergence",
        ]

    def __call__(
        self,
        model: "SentenceTransformer",
        output_path: str = None,
        epoch: int = -1,
        steps: int = -1,
    ) -> dict[str, float]:
        evaluation_message(
            epoch=epoch, steps=steps, name=self.name, truncate_dim=self.truncate_dim
        )

        with (
            nullcontext()
            if self.truncate_dim is None
            else model.truncate_sentence_embeddings(self.truncate_dim)
        ):
            queries_embeddings = torch.stack(
                model.encode(
                    self.queries,
                    batch_size=self.batch_size,
                    show_progress_bar=self.show_progress_bar,
                    convert_to_tensor=True,
                    is_query=True,
                )
            )

            documents_embeddings = torch.nn.utils.rnn.pad_sequence(
                model.encode(
                    self.documents,
                    batch_size=self.batch_size,
                    show_progress_bar=self.show_progress_bar,
                    convert_to_numpy=False,
                    is_query=False,
                ),
                batch_first=True,
                padding_value=0,
            )
        documents_embeddings = documents_embeddings.view(
            queries_embeddings.size(0), -1, *documents_embeddings.shape[1:]
        )
        scores = colbert_kd_scores(
            queries_embeddings=queries_embeddings,
            documents_embeddings=documents_embeddings,
        )
        if self.normalize_scores:
            # Compute max and min along the num_scores dimension (dim=1)
            max_scores, _ = torch.max(scores, dim=1, keepdim=True)
            min_scores, _ = torch.min(scores, dim=1, keepdim=True)

            # Avoid division by zero by adding a small epsilon
            epsilon = 1e-8

            # Normalize the scores
            scores = (scores - min_scores) / (max_scores - min_scores + epsilon)
        kl_divergence = self.loss(
            torch.nn.functional.log_softmax(scores, dim=-1),
            torch.nn.functional.log_softmax(
                torch.tensor(self.scores, device=scores.device), dim=-1
            ),
        ).item()
        metrics = self.prefix_name_to_metrics(
            {"kl_divergence": kl_divergence}, self.name
        )
        self.store_metrics_in_model_card_data(model, metrics)

        if output_path is not None and self.write_csv:
            csv_writer(
                path=os.path.join(output_path, self.csv_file),
                data=[epoch, steps, kl_divergence],
                header=self.csv_headers,
            )

        return metrics



"""Neural Cherche evaluation module for BEIR datasets."""

import random
from collections import defaultdict


def add_duplicates(queries: list[str], scores: list[list[dict]]) -> list:
    """Add back duplicates scores to the set of candidates.

    Parameters
    ----------
    queries
        List of queries.
    scores
        Scores of the retrieval model.

    """
    query_counts = defaultdict(int)
    for query in queries:
        query_counts[query] += 1

    query_to_result = {}
    for i, query in enumerate(iterable=queries):
        if query not in query_to_result:
            query_to_result[query] = scores[i]

    duplicated_scores = []
    for query in queries:
        if query in query_to_result:
            duplicated_scores.append(query_to_result[query])

    return duplicated_scores


def load_beir(dataset_name: str, split: str = "test") -> tuple[list, list, dict]:
    """Load BEIR dataset.

    Parameters
    ----------
    dataset_name
        Name of the beir dataset.
    split
        Split to load.

    Examples
    --------
    >>> from pylate import evaluation

    >>> documents, queries, qrels = evaluation.load_beir(
    ...     "scifact",
    ...     split="test",
    ... )

    >>> len(documents)
    5183

    >>> len(queries)
    300

    >>> len(qrels)
    300

    """
    from beir import util
    from beir.datasets.data_loader import GenericDataLoader

    data_path = util.download_and_unzip(
        url=f"https://public.ukp.informatik.tu-darmstadt.de/thakur/BEIR/datasets/{dataset_name}.zip",
        out_dir="./evaluation_datasets/",
    )

    documents, queries, qrels = GenericDataLoader(data_folder=data_path).load(
        split=split
    )

    documents = [
        {
            "id": document_id,
            "text": f"{document['title']} {document['text']}".strip()
            if "title" in document
            else document["text"].strip(),
        }
        for document_id, document in documents.items()
    ]

    qrels = {
        queries[query_id]: query_documents
        for query_id, query_documents in qrels.items()
    }

    return documents, list(qrels.keys()), qrels


def get_beir_triples(
    documents: list,
    queries: list[str],
    qrels: dict,
) -> list:
    """Build BEIR triples.

    Parameters
    ----------
    key
        Key.
    on
        Fields to use.
    documents
        Documents.
    queries
        Queries.

    Examples
    --------
    >>> from pylate import evaluation

    >>> documents, queries, qrels = evaluation.load_beir(
    ...     "scifact",
    ...     split="test",
    ... )

    >>> triples = evaluation.get_beir_triples(
    ...     documents=documents,
    ...     queries=queries,
    ...     qrels=qrels
    ... )

    >>> len(triples)
    339

    """

    mapping_documents = {document["id"]: document["text"] for document in documents}

    X = []
    for query, (_, query_documents) in zip(queries, qrels.items()):
        for query_document in list(query_documents.keys()):
            # Building triples, query, positive document, random negative document
            X.append(
                (
                    query,
                    mapping_documents[query_document],
                    random.choice(seq=list(mapping_documents.values())),
                )
            )
    return X


def evaluate(
    scores: list[list[dict]],
    qrels: dict,
    queries: list[str],
    metrics: list | None = None,
) -> dict[str, float]:
    """Evaluate candidates matchs.

    Parameters
    ----------
    matchs
        Matchs.
    qrels
        Qrels.
    queries
        index of queries of qrels.
    k
        Number of documents to retrieve.
    metrics
        Metrics to compute.

    Examples
    --------
    >>> from pylate import evaluation

    >>> scores = [
    ...     [{"id": "1", "score": 0.9}, {"id": "2", "score": 0.8}],
    ...     [{"id": "3", "score": 0.7}, {"id": "4", "score": 0.6}],
    ... ]

    >>> qrels = {
    ...     "query1": {"1": True, "2": True},
    ...     "query2": {"3": True, "4": True},
    ... }

    >>> queries = ["query1", "query2"]

    >>> results = evaluation.evaluate(
    ...     scores=scores,
    ...     qrels=qrels,
    ...     queries=queries,
    ...     metrics=["ndcg@10", "hits@1"],
    ... )

    """
    from ranx import Qrels, Run, evaluate

    if len(queries) > len(scores):
        scores = add_duplicates(queries=queries, scores=scores)

    qrels = Qrels(qrels=qrels)

    run_dict = {
        query: {
            match["id"]: match["score"]
            for rank, match in enumerate(iterable=query_matchs)
        }
        for query, query_matchs in zip(queries, scores)
    }

    run = Run(run=run_dict)

    if not metrics:
        metrics = ["ndcg@10"] + [f"hits@{k}" for k in [1, 2, 3, 4, 5, 10]]

    return evaluate(
        qrels=qrels,
        run=run,
        metrics=metrics,
        make_comparable=True,
    )



import numpy as np
import torch


def convert_to_tensor(
    x: torch.Tensor | np.ndarray | list[torch.Tensor | np.ndarray | list | float],
) -> torch.Tensor:
    """Converts a list or numpy array to a torch tensor.

    Parameters
    ----------
    x
        The input data. It can be a torch tensor, a numpy array, or a list of torch tensors, numpy arrays, or lists.

    Examples
    --------
    >>> import numpy as np
    >>> import torch

    >>> x = torch.tensor([[1., 1., 1.], [2., 2., 2.]])
    >>> convert_to_tensor(x)
    tensor([[1., 1., 1.],
            [2., 2., 2.]])

    >>> x = np.array([[1., 1., 1.], [2., 2., 2.]], dtype=np.float32)
    >>> convert_to_tensor(x)
    tensor([[1., 1., 1.],
            [2., 2., 2.]])

    >>> x = []
    >>> convert_to_tensor(x)
    tensor([])

    >>> x = [np.array([1., 1., 1.])]
    >>> convert_to_tensor(x)
    tensor([[1., 1., 1.]])

    >>> x = [[1., 1., 1.]]
    >>> convert_to_tensor(x)
    tensor([[1., 1., 1.]])

    >>> x = [torch.tensor([1., 1., 1.]), torch.tensor([2., 2., 2.])]
    >>> convert_to_tensor(x)
    tensor([[1., 1., 1.],
            [2., 2., 2.]])

    >>> x = np.array([], dtype=np.float32)
    >>> convert_to_tensor(x)
    tensor([])

    """
    if isinstance(x, torch.Tensor):
        return x

    if isinstance(x, np.ndarray):
        return torch.from_numpy(x)

    if isinstance(x, list):
        if not x:
            return torch.tensor([], dtype=torch.float32)

        if isinstance(x[0], np.ndarray):
            return torch.from_numpy(np.array(x, dtype=np.float32))

        if isinstance(x[0], list):
            return torch.tensor(x, dtype=torch.float32)

        if isinstance(x[0], torch.Tensor):
            return torch.stack(x)



from .collator import ColBERTCollator
from .huggingface_models import HUGGINGFACE_MODELS
from .iter_batch import iter_batch
from .multi_process import _start_multi_process_pool
from .processing import KDProcessing
from .tensor import convert_to_tensor

__all__ = [
    "HUGGINGFACE_MODELS",
    "iter_batch",
    "convert_to_tensor",
    "ColBERTCollator",
    "KDProcessing",
    "_start_multi_process_pool",
]



import copy
import json
import logging
import math
import os
import string
import traceback
from pathlib import Path
from typing import Iterable, Literal, Optional

import numpy as np
import torch
from numpy import ndarray
from scipy.cluster import hierarchy
from sentence_transformers import SentenceTransformer
from sentence_transformers.model_card import (
    SentenceTransformerModelCardData,
    generate_model_card,
)
from sentence_transformers.models import Dense as DenseSentenceTransformer
from sentence_transformers.models import Transformer
from sentence_transformers.quantization import quantize_embeddings
from sentence_transformers.similarity_functions import SimilarityFunction
from sentence_transformers.util import batch_to_device, load_file_path
from torch import nn
from tqdm.autonotebook import trange

from ..utils import _start_multi_process_pool
from .Dense import Dense

logger = logging.getLogger(__name__)


__version__ = "3.0.0"
__MODEL_HUB_ORGANIZATION__ = "sentence-transformers"


class ColBERT(SentenceTransformer):
    """
    Loads or creates a ColBERT model that can be used to map sentences / text to multi-vectors embeddings.

    Parameters
    ----------
    model_name_or_path
        If it is a filepath on disc, it loads the model from that path. If it is not a path, it first tries to download
        a pre-trained SentenceTransformer model. If that fails, tries to construct a model from the Hugging Face Hub
        with that name.
    modules
        A list of torch Modules that should be called sequentially, can be used to create custom SentenceTransformer
        models from scratch.
    device
        Device (like "cuda", "cpu", "mps", "npu") that should be used for computation. If None, checks if a GPU
        can be used.
    prompts
        A dictionary with prompts for the model. The key is the prompt name, the value is the prompt text. The prompt
        text will be prepended before any text to encode. For example:
        `{"query": "query: ", "passage": "passage: "}` or `{"clustering": "Identify the main category based on the
        titles in "}`.
    default_prompt_name
        The name of the prompt that should be used by default. If not set, no prompt will be applied.
    similarity_fn_name
        The name of the similarity function to use. Valid options are "cosine", "dot", "euclidean", and "manhattan".
        If not set, it is automatically set to "cosine" if `similarity` or `similarity_pairwise` are called while
        `model.similarity_fn_name` is still `None`.
    cache_folder
        Path to store models. Can also be set by the SENTENCE_TRANSFORMERS_HOME environment variable.
    trust_remote_code
        Whether or not to allow for custom models defined on the Hub in their own modeling files. This option should
        only be set to True for repositories you trust and in which you have read the code, as it will execute code
        present on the Hub on your local machine.
    revision
        The specific model version to use. It can be a branch name, a tag name, or a commit id, for a stored model on
        Hugging Face.
    local_files_only
        Whether or not to only look at local files (i.e., do not try to download the model).
    token
        Hugging Face authentication token to download private models.
    use_auth_token
        Deprecated argument. Please use `token` instead.
    truncate_dim
        The dimension to truncate sentence embeddings to. `None` does no truncation. Truncation is only applicable
        during inference when :meth:`SentenceTransformer.encode` is called.
    embedding_size
        The output size of the projection layer. Default to 128.
    query_prefix
        Prefix to add to the queries.
    document_prefix
        Prefix to add to the documents.
    add_special_tokens
        Add the prefix to the inputs.
    truncation
        Truncate the inputs to the encoder max lengths or use sliding window encoding.
    query_length
        The length of the query to truncate/pad to with mask tokens. If set, will override the config value. Default to 32.
    document_length
        The max length of the document to truncate. If set, will override the config value. Default to 180.
    attend_to_expansion_tokens
        Whether to attend to the expansion tokens in the attention layers model. If False, the original tokens will
        not only attend to the expansion tokens, only the expansion tokens will attend to the original tokens. Default
        is False (as in the original ColBERT codebase).
    skiplist_words
        A list of words to skip from the documents scoring (note that these tokens are used for encoding and are only skipped during the scoring). Default is the list of string.punctuation.
    model_kwargs : dict, optional
        Additional model configuration parameters to be passed to the Huggingface Transformers model. Particularly
        useful options are:

        - ``torch_dtype``: Override the default `torch.dtype` and load the model under a specific `dtype`. The
            different options are:

                1. ``torch.float16``, ``torch.bfloat16`` or ``torch.float``: load in a specified ``dtype``,
                ignoring the model's ``config.torch_dtype`` if one exists. If not specified - the model will get
                loaded in ``torch.float`` (fp32).

                2. ``"auto"`` - A ``torch_dtype`` entry in the ``config.json`` file of the model will be attempted
                to be used. If this entry isn't found then next check the ``dtype`` of the first weight in the
                checkpoint that's of a floating point type and use that as ``dtype``. This will load the model using
                the ``dtype`` it was saved in at the end of the training. It can't be used as an indicator of how the
                model was trained. Since it could be trained in one of half precision dtypes, but saved in fp32.
        - ``attn_implementation``: The attention implementation to use in the model (if relevant). Can be any of
            `"eager"` (manual implementation of the attention), `"sdpa"` (using `F.scaled_dot_product_attention
            <https://pytorch.org/docs/master/generated/torch.nn.functional.scaled_dot_product_attention.html>`_),
            or `"flash_attention_2"` (using `Dao-AILab/flash-attention
            <https://github.com/Dao-AILab/flash-attention>`_). By default, if available, SDPA will be used for
            torch>=2.1.1. The default is otherwise the manual `"eager"` implementation.

        See the `PreTrainedModel.from_pretrained
        <https://huggingface.co/docs/transformers/en/main_classes/model#transformers.PreTrainedModel.from_pretrained>`_
        documentation for more details.
    tokenizer_kwargs
        Additional tokenizer configuration parameters to be passed to the Huggingface Transformers tokenizer. See the
        `AutoTokenizer.from_pretrained
        <https://huggingface.co/docs/transformers/en/model_doc/auto#transformers.AutoTokenizer.from_pretrained>`_
        documentation for more details.
    config_kwargs
        Additional model configuration parameters to be passed to the Huggingface Transformers config. See the
        `AutoConfig.from_pretrained
        <https://huggingface.co/docs/transformers/en/model_doc/auto#transformers.AutoConfig.from_pretrained>`_
        documentation for more details.
    model_card_data
        A model card data object that contains information about the model. This is used to generate a model card when
        saving the model. If not set, a default model card data object is created.

    Examples
    --------
    >>> from pylate import models

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2",
    ...     device="cpu",
    ... )

    >>> embeddings = model.encode("Hello, how are you?")
    >>> assert isinstance(embeddings, np.ndarray)

    >>> embeddings = model.encode([
    ...     "Hello, how are you?",
    ...     "How is the weather today?"
    ... ])

    >>> assert len(embeddings) == 2
    >>> assert isinstance(embeddings[0], np.ndarray)
    >>> assert isinstance(embeddings[1], np.ndarray)

    >>> embeddings = model.encode([
    ...     [
    ...         "Hello, how are you?",
    ...         "How is the weather today?"
    ...     ],
    ...     [
    ...         "Hello, how are you?",
    ...         "How is the weather today?"
    ...     ],
    ... ])

    >>> assert len(embeddings) == 2

    >>> model.save_pretrained("test-model")

    >>> model = models.ColBERT("test-model")

    >>> embeddings = model.encode([
    ...     "Hello, how are you?",
    ...     "How is the weather today?"
    ... ])

    >>> assert len(embeddings) == 2
    >>> assert isinstance(embeddings[0], np.ndarray)
    >>> assert isinstance(embeddings[1], np.ndarray)

    """

    def __init__(
        self,
        model_name_or_path: str | None = None,
        modules: Optional[Iterable[nn.Module]] = None,
        device: str | None = None,
        prompts: dict[str, str] | None = None,
        default_prompt_name: str | None = None,
        similarity_fn_name: Optional[str | SimilarityFunction] = None,
        cache_folder: str | None = None,
        trust_remote_code: bool = False,
        revision: str | None = None,
        local_files_only: bool = False,
        token: bool | str | None = None,
        use_auth_token: bool | str | None = None,
        truncate_dim: int | None = None,
        embedding_size: int | None = None,
        bias: bool = False,
        query_prefix: str | None = "[Q] ",
        document_prefix: str | None = "[D] ",
        add_special_tokens: bool = True,
        truncation: bool = True,
        query_length: int | None = None,
        document_length: int | None = None,
        attend_to_expansion_tokens: bool = False,
        skiplist_words: list[str] | None = None,
        model_kwargs: dict | None = None,
        tokenizer_kwargs: dict | None = None,
        config_kwargs: dict | None = None,
        model_card_data: Optional[SentenceTransformerModelCardData] = None,
    ) -> None:
        model_kwargs = {} if model_kwargs is None else model_kwargs
        model_kwargs["add_pooling_layer"] = False

        self.query_prefix = query_prefix
        self.document_prefix = document_prefix
        self.query_length = query_length
        self.document_length = document_length
        self.attend_to_expansion_tokens = attend_to_expansion_tokens
        self.skiplist_words = skiplist_words

        super(ColBERT, self).__init__(
            model_name_or_path=model_name_or_path,
            modules=modules,
            device=device,
            prompts=prompts,
            default_prompt_name=default_prompt_name,
            similarity_fn_name=similarity_fn_name,
            cache_folder=cache_folder,
            trust_remote_code=trust_remote_code,
            revision=revision,
            local_files_only=local_files_only,
            token=token,
            use_auth_token=use_auth_token,
            truncate_dim=truncate_dim,
            model_kwargs=model_kwargs,
            tokenizer_kwargs=tokenizer_kwargs,
            config_kwargs=config_kwargs,
            model_card_data=model_card_data,
        )

        hidden_size = self[0].get_word_embedding_dimension()

        # Add a linear projection layer to the model in order to project the embeddings to the desired size.
        if len(self) < 2:
            # Add a linear projection layer to the model in order to project the embeddings to the desired size
            embedding_size = embedding_size or 128

            logger.warning(
                f"The checkpoint does not contain a linear projection layer. Adding one with output dimensions ({hidden_size}, {embedding_size})."
            )
            self.append(
                Dense(in_features=hidden_size, out_features=embedding_size, bias=bias)
            )

        elif (
            embedding_size is not None
            and self[1].get_sentence_embedding_dimension() != embedding_size
        ):
            logger.warning(
                f"The checkpoint contains a dense layer with output dimension ({hidden_size}, {self[1].get_sentence_embedding_dimension()}). Replacing it with a Dense layer with output dimensions ({hidden_size}, {embedding_size})."
            )
            self[1] = Dense(
                in_features=hidden_size, out_features=embedding_size, bias=bias
            )

        elif not isinstance(self[1], Dense):
            logger.warning(
                f"Converting the existing Dense layer from SentenceTransform with output dimensions ({hidden_size}, {self[1].get_sentence_embedding_dimension()})."
            )
            self[1] = Dense.from_sentence_transformers(dense=self[1])
        else:
            logger.warning("Pylate model loaded successfully.")

        self.to(device)
        self.is_hpu_graph_enabled = False

        self.tokenizer.add_tokens([self.query_prefix, self.document_prefix])
        self._first_module().auto_model.resize_token_embeddings(len(self.tokenizer))

        self.document_prefix_id = self.tokenizer.convert_tokens_to_ids(
            self.document_prefix
        )

        # Set the query prefix ID using the tokenizer.
        self.query_prefix_id = self.tokenizer.convert_tokens_to_ids(self.query_prefix)

        # Set the padding token ID to be the same as the mask token ID for queries.
        self.tokenizer.pad_token_id = self.tokenizer.mask_token_id

        # Override the configuration values with the provided arguments, if any.
        self.document_length = (
            document_length
            if document_length is not None
            else self.document_length or 180
        )

        self.query_length = (
            query_length if query_length is not None else self.query_length or 32
        )

        self.skiplist_words = (
            skiplist_words
            if skiplist_words is not None
            else self.skiplist_words or list(string.punctuation)
        )

        # Convert skiplist words to their corresponding token IDs.
        self.skiplist = [
            self.tokenizer.convert_tokens_to_ids(word) for word in self.skiplist_words
        ]

    @staticmethod
    def load(input_path) -> "ColBERT":
        return ColBERT(model_name_or_path=input_path)

    def __len__(self) -> int:
        return len(self._modules)

    @staticmethod
    def insert_prefix_token(input_ids: torch.Tensor, prefix_id: int) -> torch.Tensor:
        """Inserts a prefix token at the beginning of each sequence in the input tensor."""
        prefix_tensor = torch.full(
            size=(input_ids.size(dim=0), 1),
            fill_value=prefix_id,
            dtype=input_ids.dtype,
            device=input_ids.device,
        )

        return torch.cat(
            tensors=[input_ids[:, :1], prefix_tensor, input_ids[:, 1:]], dim=1
        )

    def encode(
        self,
        sentences: str | list[str],
        prompt_name: str | None = None,
        prompt: str | None = None,
        batch_size: int = 32,
        show_progress_bar: bool = None,
        precision: Literal["float32", "int8", "uint8", "binary", "ubinary"] = "float32",
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
        padding: bool = False,
        device: str = None,
        normalize_embeddings: bool = True,
        is_query: bool = True,
        pool_factor: int = 1,
        protected_tokens: int = 1,
    ) -> list[torch.Tensor] | ndarray | torch.Tensor:
        """
        Computes sentence embeddings.

        Parameters
        ----------
        sentences
            The sentences to embed.
        prompt_name
            The name of the prompt to use for encoding. Must be a key in the `prompts` dictionary, which is either set in
            the constructor or loaded from the model configuration. For example, if `prompt_name` is "query" and the
            `prompts` is {"query": "query: ", ...}, then the sentence "What is the capital of France?" will be encoded as
            "query: What is the capital of France?" because the sentence is appended to the prompt. If `prompt` is also
            set, this argument is ignored. Defaults to None.
        prompt
            The prompt to use for encoding. For example, if the prompt is "query: ", then the sentence "What is the capital
            of France?" will be encoded as "query: What is the capital of France?" because the sentence is appended to the
            prompt. If `prompt` is set, `prompt_name` is ignored. Defaults to None.
        batch_size
            The batch size used for the computation. Defaults to 32.
        show_progress_bar
            Whether to output a progress bar when encoding sentences. Defaults to None.
        output_value
            The type of embeddings to return: "sentence_embedding" to get sentence embeddings, "token_embeddings" to get
            wordpiece token embeddings, and `None` to get all output values. Defaults to "sentence_embedding".
        precision
            The precision to use for the embeddings. Can be "float32", "int8", "uint8", "binary", or "ubinary". All
            non-float32 precisions are quantized embeddings. Quantized embeddings are smaller in size and faster to compute,
            but may have lower accuracy. They are useful for reducing the size of the embeddings of a corpus for semantic
            search, among other tasks. Defaults to "float32".
        convert_to_numpy
            Whether the output should be a list of numpy vectors. If False, it is a list of PyTorch tensors. Defaults to True.
        convert_to_tensor
            Whether the output should be one large tensor. Overwrites `convert_to_numpy`. Defaults to False.
        device
            Which :class:`torch.device` to use for the computation. Defaults to None.
        normalize_embeddings
            Whether to normalize returned vectors to have length 1. In that case, the faster dot-product (util.dot_score)
            instead of cosine similarity can be used. Defaults to False.
        is_query
            Whether the input sentences are queries. If True, the query prefix is added to the input sentences and the
            sequence is padded; otherwise, the document prefix is added and the sequence is not padded. Defaults to True.
        pool_factor
            The factor by which to pool the document embeddings, resulting in 1/pool_factor of the original tokens. If set
            to 1, no pooling is done; if set to 2, 50% of the tokens are kept; if set to 3, 33%, and so on. Defaults to 1.
        protected_tokens
            The number of tokens at the beginning of the sequence that should not be pooled. Defaults to 1 (CLS token).

        """
        if isinstance(sentences, list):
            # If we have a list of list of sentences, we encode each list separately.
            if isinstance(sentences[0], list):
                embeddings = []

                for batch in sentences:
                    batch_embedings = self.encode(
                        sentences=batch,
                        prompt_name=prompt_name,
                        prompt=prompt,
                        batch_size=batch_size,
                        show_progress_bar=show_progress_bar,
                        precision=precision,
                        convert_to_numpy=convert_to_numpy,
                        convert_to_tensor=convert_to_tensor,
                        padding=padding,
                        device=device,
                        normalize_embeddings=normalize_embeddings,
                        is_query=is_query,
                        pool_factor=pool_factor,
                        protected_tokens=protected_tokens,
                    )

                    batch_embedings = (
                        torch.stack(batch_embedings)
                        if convert_to_tensor
                        else batch_embedings
                    )

                    embeddings.append(batch_embedings)

                return embeddings

        if self.device.type == "hpu" and not self.is_hpu_graph_enabled:
            import habana_frameworks.torch as ht

            ht.hpu.wrap_in_hpu_graph(self, disable_tensor_cache=True)
            self.is_hpu_graph_enabled = True

        self.eval()
        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO
                or logger.getEffectiveLevel() == logging.DEBUG
            )

        # Convert to tensor takes precedence over convert to numpy
        if not convert_to_numpy:
            convert_to_tensor = True
        convert_to_numpy = not convert_to_tensor

        # TODO: We cannot convert to tensor/numpy for token embeddings as they are not the same size
        # if output_value != "sentence_embedding":
        # convert_to_tensor = False
        # convert_to_numpy = False

        input_was_string = False
        if isinstance(sentences, str) or not hasattr(sentences, "__len__"):
            sentences = [sentences]
            input_was_string = True

        if prompt is not None and prompt_name is not None:
            logger.warning(
                "Provide either a `prompt` or a `prompt_name`, not both. "
                "Ignoring the `prompt_name` in favor of the provided `prompt`."
            )

        elif prompt is None:
            if prompt_name is not None:
                prompt = self.prompts.get(prompt_name)
                if prompt is None:
                    raise ValueError(
                        f"Prompt name '{prompt_name}' not found in the configured prompts dictionary. "
                        f"Available keys are: {list(self.prompts.keys())!r}."
                    )
            else:
                prompt = self.prompts.get(self.default_prompt_name)

        extra_features = {}
        if prompt is not None:
            sentences = [prompt + sentence for sentence in sentences]

            # Some models require removing the prompt before pooling (e.g. Instructor, Grit).
            # Tracking the prompt length allow us to remove the prompt during pooling.
            tokenized_prompt = self.tokenize([prompt])
            if "input_ids" in tokenized_prompt:
                extra_features["prompt_length"] = (
                    tokenized_prompt["input_ids"].shape[-1] - 1
                )

        if device is None:
            device = self.device

        self.to(device)

        all_embeddings = []
        length_sorted_idx = np.argsort([-self._text_length(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]

        for start_index in trange(
            0,
            len(sentences),
            batch_size,
            desc=f"Encoding queries (bs={batch_size})"
            if is_query
            else f"Encoding documents (bs={batch_size})",
            disable=not show_progress_bar,
        ):
            sentences_batch = sentences_sorted[start_index : start_index + batch_size]
            features = self.tokenize(texts=sentences_batch, is_query=is_query)

            if self.device.type == "hpu":
                if "input_ids" in features:
                    curr_tokenize_len = features["input_ids"].shape

                    additional_pad_len = (
                        2 ** math.ceil(math.log2(curr_tokenize_len[1]))
                        - curr_tokenize_len[1]
                    )

                    features["input_ids"] = torch.cat(
                        tensors=(
                            features["input_ids"],
                            torch.ones(
                                size=(curr_tokenize_len[0], additional_pad_len),
                                dtype=torch.int8,
                            ),
                        ),
                        dim=-1,
                    )

                    features["attention_mask"] = torch.cat(
                        tensors=(
                            features["attention_mask"],
                            torch.zeros(
                                size=(curr_tokenize_len[0], additional_pad_len),
                                dtype=torch.int8,
                            ),
                        ),
                        dim=-1,
                    )

                    if "token_type_ids" in features:
                        features["token_type_ids"] = torch.cat(
                            tensors=(
                                features["token_type_ids"],
                                torch.zeros(
                                    size=(curr_tokenize_len[0], additional_pad_len),
                                    dtype=torch.int8,
                                ),
                            ),
                            dim=-1,
                        )

            features = batch_to_device(batch=features, target_device=device)
            features.update(extra_features)

            with torch.no_grad():
                # TODO: add the truncate/sliding window logic here
                out_features = self.forward(input=features)
                if self.device.type == "hpu":
                    out_features = copy.deepcopy(out_features)

                if not is_query:
                    # Compute the mask for the skiplist (punctuation symbols)
                    skiplist_mask = self.skiplist_mask(
                        input_ids=features["input_ids"], skiplist=self.skiplist
                    )
                    masks = torch.logical_and(
                        input=skiplist_mask, other=out_features["attention_mask"]
                    )
                else:
                    # We keep all tokens in the query (no skiplist) and we do not want to prune expansion tokens in queries even if we do not attend to them in attention layers
                    masks = torch.ones_like(
                        input=out_features["input_ids"], dtype=torch.bool
                    )

                embeddings = []
                for (
                    token_embedding,
                    mask,
                ) in zip(out_features["token_embeddings"], masks):
                    token_embedding = (
                        torch.nn.functional.normalize(
                            input=token_embedding[mask], p=2, dim=1
                        )
                        if normalize_embeddings
                        else token_embedding[mask]
                    )
                    embeddings.append(token_embedding)

                # Pool factor must be greater than 1: keeping 1 over pool_factor tokens embeddings.
                if pool_factor > 1 and not is_query:
                    embeddings = self.pool_embeddings_hierarchical(
                        documents_embeddings=embeddings,
                        pool_factor=pool_factor,
                        protected_tokens=protected_tokens,
                    )

                # fixes for #522 and #487 to avoid oom problems on gpu with large datasets
                if convert_to_numpy:
                    embeddings = [embedding.cpu() for embedding in embeddings]

                all_embeddings.extend(embeddings)

        # Pad the embeddings to the same length. Documents can have different lengths while queries are already padded.
        if padding:
            all_embeddings = torch.nn.utils.rnn.pad_sequence(
                sequences=all_embeddings, batch_first=True, padding_value=0
            )

            # Create a list of tensors.
            all_embeddings = torch.split(
                tensor=all_embeddings, split_size_or_sections=1, dim=0
            )

        all_embeddings = [all_embeddings[idx] for idx in np.argsort(length_sorted_idx)]

        if precision and precision != "float32":
            all_embeddings = quantize_embeddings(
                embeddings=all_embeddings, precision=precision
            )

        # Return a list of arrays instead of single contiguous array since documents can have different lengths.
        if convert_to_tensor:
            if not len(all_embeddings):
                return torch.tensor()

            if isinstance(all_embeddings, np.ndarray):
                all_embeddings = [
                    torch.from_numpy(ndarray=embedding) for embedding in all_embeddings
                ]

        elif convert_to_numpy:
            bloat = all_embeddings[0].dtype == torch.bfloat16
            all_embeddings = [
                embedding.float().numpy() if bloat else embedding.numpy()
                for embedding in all_embeddings
            ]

        return all_embeddings[0] if input_was_string else all_embeddings

    def pool_embeddings_hierarchical(
        self,
        documents_embeddings: list[torch.Tensor],
        pool_factor: int = 1,
        protected_tokens: int = 1,
    ) -> list[torch.Tensor]:
        """
        Pools the embeddings hierarchically by clustering and averaging them.

        Parameters
        ----------
        document_embeddings_list
            A list of embeddings for each document.
        pool_factor
            Factor to determine the number of clusters. Defaults to 1.
        protected_tokens
            Number of tokens to protect from pooling at the start of each document. Defaults to 1.

        Returns
        -------
            A list of pooled embeddings for each document.
        """
        device = torch.device(device="cuda" if torch.cuda.is_available() else "cpu")
        pooled_embeddings = []

        for document_embeddings in documents_embeddings:
            document_embeddings = document_embeddings.to(device=device)

            # Separate protected tokens from the rest
            protected_embeddings = document_embeddings[:protected_tokens]
            embeddings_to_pool = document_embeddings[protected_tokens:]

            # Compute cosine similarity and convert to distance matrix
            cosine_similarities = torch.mm(
                input=embeddings_to_pool, mat2=embeddings_to_pool.t()
            )
            distance_matrix = 1 - cosine_similarities.cpu().numpy()

            # Perform hierarchical clustering using Ward's method
            clusters = hierarchy.linkage(distance_matrix, method="ward")
            num_embeddings = len(embeddings_to_pool)

            # Determine the number of clusters based on pool_factor
            num_clusters = max(num_embeddings // pool_factor, 1)
            cluster_labels = hierarchy.fcluster(
                clusters, t=num_clusters, criterion="maxclust"
            )

            # Pool embeddings within each cluster
            pooled_document_embeddings = []
            for cluster_id in range(1, num_clusters + 1):
                cluster_indices = torch.where(
                    condition=torch.tensor(
                        data=cluster_labels == cluster_id, device=device
                    )
                )[0]
                if cluster_indices.numel() > 0:
                    cluster_embedding = embeddings_to_pool[cluster_indices].mean(dim=0)
                    pooled_document_embeddings.append(cluster_embedding)

            # Re-append protected embeddings
            pooled_document_embeddings.extend(protected_embeddings)
            pooled_embeddings.append(torch.stack(tensors=pooled_document_embeddings))

        return pooled_embeddings

    @staticmethod
    def skiplist_mask(input_ids: torch.Tensor, skiplist: list[int]) -> torch.Tensor:
        """Create a mask for the set of input_ids that are in the skiplist."""
        skiplist = torch.tensor(
            data=skiplist, dtype=torch.long, device=input_ids.device
        )

        # Create a tensor of ones with the same shape as input_ids.
        mask = torch.ones_like(input=input_ids, dtype=torch.bool)

        # Update the mask for each token in the skiplist.
        for token_id in skiplist:
            mask = torch.where(
                condition=input_ids == token_id,
                input=torch.tensor(data=0, dtype=torch.bool, device=input_ids.device),
                other=mask,
            )

        return mask

    def start_multi_process_pool(self, target_devices: list[str] = None) -> dict:
        """Starts a multi-process pool to process the encoding with several independent processes.
        This method is recommended if you want to encode on multiple GPUs or CPUs. It is advised
        to start only one process per GPU. This method works together with encode_multi_process
        and stop_multi_process_pool.

        Parameters
        ----------
        target_devices
            PyTorch target devices, e.g. ["cuda:0", "cuda:1", ...], ["npu:0", "npu:1", ...], or ["cpu", "cpu", "cpu", "cpu"].
            If target_devices is None and CUDA/NPU is available, then all available CUDA/NPU devices will be used. If target_devices is None and
            CUDA/NPU is not available, then 4 CPU devices will be used.

        Returns
        -------
            A dictionary with the target processes, an input queue, and an output queue.
        """
        return _start_multi_process_pool(model=self, target_devices=target_devices)

    def encode_multi_process(
        self,
        sentences: list[str],
        pool: dict[str, object],
        prompt_name: str | None = None,
        prompt: str | None = None,
        batch_size: int = 32,
        chunk_size: int = None,
        precision: Literal["float32", "int8", "uint8", "binary", "ubinary"] = "float32",
        normalize_embeddings: bool = True,
        padding: bool = False,
        is_query: bool = True,
        pool_factor: int = 1,
        protected_tokens: int = 1,
    ) -> list[np.ndarray]:
        """
        Encodes a list of sentences using multiple processes and GPUs via
        :meth:`SentenceTransformer.encode <sentence_transformers.SentenceTransformer.encode>`.
        The sentences are chunked into smaller packages and sent to individual processes, which encode them on different
        GPUs or CPUs. This method is only suitable for encoding large sets of sentences.

        Parameters
        ----------
        sentences
            List of sentences to encode.
        pool
            A pool of workers started with SentenceTransformer.start_multi_process_pool.
        prompt_name
            The name of the prompt to use for encoding. Must be a key in the `prompts` dictionary,
            which is either set in the constructor or loaded from the model configuration. For example if
            ``prompt_name`` is "query" and the ``prompts`` is {"query": "query: ", ...}, then the sentence "What
            is the capital of France?" will be encoded as "query: What is the capital of France?" because the sentence
            is appended to the prompt. If ``prompt`` is also set, this argument is ignored. Defaults to None.
        prompt
            The prompt to use for encoding. For example, if the prompt is "query: ", then the
            sentence "What is the capital of France?" will be encoded as "query: What is the capital of France?"
            because the sentence is appended to the prompt. If ``prompt`` is set, ``prompt_name`` is ignored. Defaults to None.
        batch_size
            Encode sentences with batch size. (default: 32)
        chunk_size
            Sentences are chunked and sent to the individual processes. If None, it determines a
            sensible size. Defaults to None.
        precision
            The precision to use for the
            embeddings. Can be "float32", "int8", "uint8", "binary", or "ubinary". All non-float32 precisions
            are quantized embeddings. Quantized embeddings are smaller in size and faster to compute, but may
            have lower accuracy. They are useful for reducing the size of the embeddings of a corpus for
            semantic search, among other tasks. Defaults to "float32".
        normalize_embeddings
            Whether to normalize returned vectors to have length 1. In that case,
            the faster dot-product (util.dot_score) instead of cosine similarity can be used. Defaults to True.

        Examples
        --------
        >>> from pylate import models

        >>> model = models.ColBERT(
        ...     "sentence-transformers/all-MiniLM-L6-v2",
        ...     device="cpu"
        ... )

        >>> pool = model.start_multi_process_pool()
        >>> embeddings = model.encode_multi_process(
        ...     sentences=["The weather is lovely today.", "It's so sunny outside!", "He drove to the stadium."],
        ...     pool=pool,
        ...     batch_size=1,
        ...     is_query=True,
        ... )
        >>> model.stop_multi_process_pool(pool)

        >>> assert len(embeddings) == 3

        """

        if chunk_size is None:
            chunk_size = min(
                math.ceil(len(sentences) / len(pool["processes"]) / 10), 5000
            )

        logger.debug(
            f"Chunk data into {math.ceil(len(sentences) / chunk_size)} packages of size {chunk_size}"
        )

        input_queue = pool["input"]
        last_chunk_id = 0
        chunk = []

        for sentence in sentences:
            chunk.append(sentence)
            if len(chunk) >= chunk_size:
                input_queue.put(
                    [
                        last_chunk_id,
                        batch_size,
                        chunk,
                        prompt_name,
                        prompt,
                        precision,
                        normalize_embeddings,
                        padding,
                        is_query,
                        pool_factor,
                        protected_tokens,
                    ]
                )
                last_chunk_id += 1
                chunk = []

        if len(chunk) > 0:
            input_queue.put(
                [
                    last_chunk_id,
                    batch_size,
                    chunk,
                    prompt_name,
                    prompt,
                    precision,
                    normalize_embeddings,
                    padding,
                    is_query,
                    pool_factor,
                    protected_tokens,
                ]
            )
            last_chunk_id += 1

        output_queue = pool["output"]
        results_list = sorted(
            [output_queue.get() for _ in range(last_chunk_id)], key=lambda x: x[0]
        )
        return [np.concatenate(result[1]) for result in results_list]

    def tokenize(
        self,
        texts: list[str] | list[dict] | list[tuple[str, str]],
        is_query: bool = True,
        pad_document: bool = False,
    ) -> dict[str, torch.Tensor]:
        """
        Tokenizes the input texts.

        Args:
            texts (Union[list[str], list[dict], list[tuple[str, str]]]): A list of texts to be tokenized.
            is_query (bool): Flag to indicate if the texts are queries. Defaults to True.
            pad_document (bool): Flag to indicate if documents should be padded to max length. Defaults to False.

        Returns:
            dict[str, torch.Tensor]: A dictionary of tensors with the tokenized texts, including "input_ids",
                "attention_mask", and optionally "token_type_ids".
        """
        # Set max sequence length based on whether the input is a query or document
        max_length = self.query_length if is_query else self.document_length
        self._first_module().max_seq_length = (
            max_length - 1
        )  # Subtract 1 for the prefix token

        # Pad queries (query expansion) and handle padding for documents if specified
        tokenize_args = {"padding": "max_length"} if pad_document or is_query else {}

        # Tokenize the texts
        tokenized_outputs = self._first_module().tokenize(texts, **tokenize_args)

        # Determine prefix ID based on input type
        prefix_id = self.query_prefix_id if is_query else self.document_prefix_id

        # Insert prefix token and update attention mask
        tokenized_outputs["input_ids"] = self.insert_prefix_token(
            tokenized_outputs["input_ids"], prefix_id
        )
        tokenized_outputs["attention_mask"] = self.insert_prefix_token(
            tokenized_outputs["attention_mask"], 1
        )

        # Update token type IDs if they exist
        if "token_type_ids" in tokenized_outputs:
            tokenized_outputs["token_type_ids"] = self.insert_prefix_token(
                tokenized_outputs["token_type_ids"], 0
            )

        # Adjust attention mask for expansion tokens if required
        if is_query and self.attend_to_expansion_tokens:
            tokenized_outputs["attention_mask"].fill_(1)

        return tokenized_outputs

    def save(
        self,
        path: str,
        model_name: str | None = None,
        create_model_card: bool = True,
        train_datasets: list[str] | None = None,
        safe_serialization: bool = True,
    ) -> None:
        """
        Saves a model and its configuration files to a directory, so that it can be loaded
        with ``SentenceTransformer(path)`` again.

        Args:
            path (str): Path on disc where the model will be saved.
            model_name (str, optional): Optional model name.
            create_model_card (bool, optional): If True, create a README.md with basic information about this model.
            train_datasets (list[str], optional): Optional list with the names of the datasets used to train the model.
            safe_serialization (bool, optional): If True, save the model using safetensors. If False, save the model
                the traditional (but unsafe) PyTorch way.
        """
        super().save(
            path,
            model_name=model_name,
            create_model_card=create_model_card,
            train_datasets=train_datasets,
            safe_serialization=safe_serialization,
        )

        with open(os.path.join(path, "config_sentence_transformers.json"), "w") as fOut:
            config = self._model_config.copy()
            config["prompts"] = self.prompts
            config["default_prompt_name"] = self.default_prompt_name
            config["similarity_fn_name"] = self.similarity_fn_name
            config["query_prefix"] = self.query_prefix
            config["document_prefix"] = self.document_prefix
            config["query_length"] = self.query_length
            config["document_length"] = self.document_length
            config["attend_to_expansion_tokens"] = self.attend_to_expansion_tokens
            config["skiplist_words"] = self.skiplist_words
            json.dump(config, fOut, indent=2)

    def _create_model_card(
        self,
        path: str,
        model_name: str | None = None,
        train_datasets: list[str] | None = None,
    ) -> None:
        """
        Create an automatic model and stores it in the specified path. If no training was done and the loaded model
        was a Sentence Transformer model already, then its model card is reused.

        Parameters
        ----------
        path
            The path where the model card will be stored.
        model_name
            The name of the model.
        """
        if model_name:
            model_path = Path(model_name)
            if not model_path.exists() and not self.model_card_data.model_id:
                self.model_card_data.model_id = model_name

        # If we loaded a Sentence Transformer model from the Hub, and no training was done, then
        # we don't generate a new model card, but reuse the old one instead.
        if self._model_card_text and self.model_card_data.trainer is None:
            model_card = self._model_card_text
        else:
            try:
                model_card = generate_model_card(model=self)
            except Exception:
                logger.error(
                    f"Error while generating model card:\n{traceback.format_exc()}"
                    "Consider opening an issue on https://github.com/UKPLab/sentence-transformers/issues with this traceback.\n"
                    "Skipping model card creation."
                )
                return

        with open(os.path.join(path, "README.md"), "w", encoding="utf8") as fOut:
            fOut.write(model_card)

    def _load_auto_model(
        self,
        model_name_or_path: str,
        token: bool | str | None,
        cache_folder: str | None,
        revision: str | None = None,
        trust_remote_code: bool = False,
        local_files_only: bool = False,
        model_kwargs: dict | None = None,
        tokenizer_kwargs: dict | None = None,
        config_kwargs: dict | None = None,
    ) -> list[nn.Module]:
        """Create a Transformer model from a model name or path. This module is distinct
        from SentenceTransformer as it do not set the pooling layer.

        Parameters
        ----------
        model_name_or_path
            The name or path of the pre-trained model.
        token
            The token to use for the model.
        cache_folder
            The folder to cache the model.
        revision
            The revision of the model. Defaults to None.
        trust_remote_code
            Whether to trust remote code. Defaults to False.
        local_files_only
            Whether to use only local files. Defaults to False.
        model_kwargs
            Additional keyword arguments for the model. Defaults to None.
        tokenizer_kwargs
            Additional keyword arguments for the tokenizer. Defaults to None.
        config_kwargs
            Additional keyword arguments for the config. Defaults to None.

        """
        logger.warning(
            f"No sentence-transformers model found with name {model_name_or_path}. Creating a ColBERT model from base encoder."
        )

        shared_kwargs = {
            "token": token,
            "trust_remote_code": trust_remote_code,
            "revision": revision,
            "local_files_only": local_files_only,
        }

        model_kwargs = (
            shared_kwargs if model_kwargs is None else {**shared_kwargs, **model_kwargs}
        )

        tokenizer_kwargs = (
            shared_kwargs
            if tokenizer_kwargs is None
            else {**shared_kwargs, **tokenizer_kwargs}
        )

        config_kwargs = (
            shared_kwargs
            if config_kwargs is None
            else {**shared_kwargs, **config_kwargs}
        )

        transformer_model = Transformer(
            model_name_or_path=model_name_or_path,
            cache_dir=cache_folder,
            model_args=model_kwargs,
            tokenizer_args=tokenizer_kwargs,
            config_args=config_kwargs,
        )

        self.model_card_data.set_base_model(
            model_id=model_name_or_path, revision=revision
        )

        return [transformer_model]

    def _load_sbert_model(
        self,
        model_name_or_path: str,
        token: bool | str | None,
        cache_folder: str | None,
        revision: str | None = None,
        trust_remote_code: bool = False,
        local_files_only: bool = False,
        model_kwargs: dict | None = None,
        tokenizer_kwargs: dict | None = None,
        config_kwargs: dict | None = None,
    ) -> list[nn.Module]:
        """Create a Sentence Transformer model from a model name or path."""
        modules = super()._load_sbert_model(
            model_name_or_path=model_name_or_path,
            token=token,
            cache_folder=cache_folder,
            revision=revision,
            trust_remote_code=trust_remote_code,
            local_files_only=local_files_only,
            model_kwargs=model_kwargs,
            tokenizer_kwargs=tokenizer_kwargs,
            config_kwargs=config_kwargs,
        )

        config_sentence_transformers_json_path = load_file_path(
            model_name_or_path=model_name_or_path,
            filename="config_sentence_transformers.json",
            token=token,
            cache_folder=cache_folder,
            revision=revision,
            local_files_only=local_files_only,
        )

        if config_sentence_transformers_json_path is not None:
            with open(file=config_sentence_transformers_json_path) as fIn:
                self._model_config = json.load(fp=fIn)

            # Loading the query/document prefixes and query_length
            if "query_prefix" in self._model_config:
                self.query_prefix = self._model_config["query_prefix"]
            if "document_prefix" in self._model_config:
                self.document_prefix = self._model_config["document_prefix"]
            if "query_length" in self._model_config:
                self.query_length = self._model_config["query_length"]
            if "document_length" in self._model_config:
                self.document_length = self._model_config["document_length"]
            if "attend_to_expansion_tokens" in self._model_config:
                self.attend_to_expansion_tokens = self._model_config[
                    "attend_to_expansion_tokens"
                ]
            if "skiplist_words" in self._model_config:
                self.skiplist_words = self._model_config["skiplist_words"]

        return [
            module
            for module in modules.values()
            if isinstance(module, Transformer)
            or isinstance(module, DenseSentenceTransformer)
        ]



import csv
import logging
import os
from contextlib import nullcontext

from sentence_transformers.evaluation import TripletEvaluator

from ..models import ColBERT
from ..scores import colbert_scores_pairwise

logger = logging.getLogger(__name__)


logger = logging.getLogger(__name__)


def csv_writer(
    path: str | None,
    data: list[str | float | int],
    header: list[str],
) -> None:
    """Write the results to a CSV file.

    Parameters
    ----------
    path
        The path to the CSV file.
    rows
        The rows to write to the CSV file.
    header
        The header of the CSV file.

    Examples
    --------
    >>> import pandas as pd

    >>> csv_writer(
    ...     path="results.csv",
    ...     data=[1, 2, 0.8],
    ...     header=["epoch", "steps", "accuracy"],
    ... )

    >>> df = pd.read_csv("results.csv")

    >>> assert df.columns.tolist() == ["epoch", "steps", "accuracy"]

    """
    mode = "w" if not os.path.isfile(path=path) else "a"
    with open(file=path, newline="", mode=mode, encoding="utf-8") as f:
        writer = csv.writer(f)
        if mode == "w":
            writer.writerow(header)
        writer.writerow(data)


def evaluation_message(
    epoch: int,
    steps: int,
    name: str,
    truncate_dim: int | None = None,
) -> None:
    """Prints an evaluation message."""
    out_txt = ""
    if epoch != -1:
        out_txt = (
            f" after epoch {epoch}"
            if steps == -1
            else f" in epoch {epoch} after {steps} steps"
        )

    if truncate_dim is not None:
        out_txt += f" (truncated to {truncate_dim})"

    logger.info(
        f"TripletEvaluator: Evaluating the model on the {name} dataset{out_txt}:"
    )


class ColBERTTripletEvaluator(TripletEvaluator):
    """Evaluate a model based on a set of triples. The evaluation will compare the
    score between the anchor and the positive sample with the score between the anchor
    and the negative sample. The accuracy is computed as the number of times the score
    between the anchor and the positive sample is higher than the score between the anchor
    and the negative sample.

    Parameters
    ----------
    anchors
        Sentences to check similarity to. (e.g. a query)
    positives
        List of positive sentences
    negatives
        List of negative sentences
    name
        Name for the output.
    batch_size
        Batch size used to compute embeddings.
    show_progress_bar
        If true, prints a progress bar.
    write_csv
        Wether or not to write results to a CSV file.
    truncate_dim
        The dimension to truncate sentence embeddings to. If None, do not truncate.

    Examples
    --------
    >>> from pylate import evaluation, models

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2",
    ...     device="cpu",
    ... )

    >>> anchors = [
    ...     "fruits are healthy.",
    ...     "fruits are healthy.",
    ... ]

    >>> positives = [
    ...     "fruits are good for health.",
    ...     "Fruits are growing in the trees.",
    ... ]

    >>> negatives = [
    ...     "Fruits are growing in the trees.",
    ...     "fruits are good for health.",
    ... ]

    >>> triplet_evaluation = evaluation.ColBERTTripletEvaluator(
    ...     anchors=anchors,
    ...     positives=positives,
    ...     negatives=negatives,
    ...     write_csv=True,
    ... )

    >>> results = triplet_evaluation(model=model, output_path=".")

    >>> results
    {'accuracy': 0.5}

    >>> triplet_evaluation.csv_headers
    ['epoch', 'steps', 'accuracy']

    >>> import pandas as pd
    >>> df = pd.read_csv(triplet_evaluation.csv_file)
    >>> assert df.columns.tolist() == triplet_evaluation.csv_headers

    """

    def __init__(
        self,
        anchors: list[str],
        positives: list[str],
        negatives: list[str],
        name: str = "",
        batch_size: int = 32,
        show_progress_bar: bool = False,
        write_csv: bool = True,
        truncate_dim: int | None = None,
    ) -> None:
        super().__init__(
            anchors=anchors,
            positives=positives,
            negatives=negatives,
            main_distance_function=None,
            name=name,
            batch_size=batch_size,
            show_progress_bar=show_progress_bar,
            write_csv=write_csv,
            truncate_dim=truncate_dim,
        )

        self.csv_headers = [
            "epoch",
            "steps",
            "accuracy",
        ]

        self.metrics = [
            "accuracy",
        ]

        self.primary_metric = "accuracy"

    def __call__(
        self,
        model: ColBERT,
        output_path: str = None,
        epoch: int = -1,
        steps: int = -1,
    ) -> dict[str, float]:
        """Evaluate the model on the triplet dataset. Measure the scoring between the anchor
        and the positive with every other positive and negative samples using HITS@K.
        """
        evaluation_message(
            epoch=epoch, steps=steps, name=self.name, truncate_dim=self.truncate_dim
        )

        with (
            nullcontext()
            if self.truncate_dim is None
            else model.truncate_sentence_embeddings(truncate_dim=self.truncate_dim)
        ):
            embeddings_anchors = model.encode(
                sentences=self.anchors,
                batch_size=self.batch_size,
                show_progress_bar=self.show_progress_bar,
                convert_to_tensor=True,
                is_query=True,
            )
            embeddings_positives = model.encode(
                sentences=self.positives,
                batch_size=self.batch_size,
                show_progress_bar=self.show_progress_bar,
                convert_to_numpy=False,
                is_query=False,
            )
            embeddings_negatives = model.encode(
                sentences=self.negatives,
                batch_size=self.batch_size,
                show_progress_bar=self.show_progress_bar,
                convert_to_numpy=False,
                is_query=False,
            )

        # Colbert pairwise scores
        positive_scores = colbert_scores_pairwise(
            queries_embeddings=embeddings_anchors,
            documents_embeddings=embeddings_positives,
        )

        negative_scores = colbert_scores_pairwise(
            queries_embeddings=embeddings_anchors,
            documents_embeddings=embeddings_negatives,
        )

        metrics = {
            "accuracy": (
                sum(positive_scores > negative_scores) / len(positive_scores)
            ).item()
        }

        for metric in self.metrics:
            logger.info(f"{metric.capitalize()}: \t{metrics[metric]:.2f}")

        self.store_metrics_in_model_card_data(model=model, metrics=metrics)

        if output_path is not None and self.write_csv:
            csv_writer(
                path=os.path.join(output_path, self.csv_file),
                header=self.csv_headers,
                data=[
                    epoch,
                    steps,
                ]
                + [metrics[metric] for metric in self.metrics],
            )

        return metrics



from typing import Iterable

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from ..models import ColBERT
from ..scores import colbert_scores


def extract_skiplist_mask(
    sentence_features: Iterable[dict[str, torch.Tensor]],
    skiplist: list[int],
) -> list[torch.Tensor]:
    """Extracts the attention masks from the sentence features. We apply a skiplist mask to the documents.
    We skip the first sentence feature because it is the query.

    Examples
    --------
    >>> import torch

    >>> sentence_features = [
    ...     {
    ...         "input_ids": torch.tensor([[1, 2, 3, 4]]),
    ...         "attention_mask": torch.tensor([[1, 1, 1, 1]]),
    ...     },
    ...     {
    ...         "input_ids": torch.tensor([[1, 2, 3, 4]]),
    ...         "attention_mask": torch.tensor([[1, 1, 1, 1]]),
    ...     },
    ...     {
    ...         "input_ids": torch.tensor([[1, 2, 3, 4]]),
    ...         "attention_mask": torch.tensor([[1, 1, 1, 1]]),
    ...     },
    ... ]

    >>> extract_skiplist_mask(
    ...     sentence_features=sentence_features,
    ...     skiplist=[1, 2, 3],
    ... )
    [tensor([[True, True, True, True]]), tensor([[False, False, False,  True]]), tensor([[False, False, False,  True]])]

    """
    attention_masks = [
        sentence_feature["attention_mask"] for sentence_feature in sentence_features
    ]

    skiplist_masks = [
        torch.ones_like(sentence_features[0]["input_ids"], dtype=torch.bool)
    ]

    # We skip the first sentence feature because it is the query.
    skiplist_masks.extend(
        [
            ColBERT.skiplist_mask(
                input_ids=sentence_feature["input_ids"], skiplist=skiplist
            )
            for sentence_feature in sentence_features[1:]
        ]
    )

    return [
        torch.logical_and(skiplist_mask, attention_mask)
        for skiplist_mask, attention_mask in zip(skiplist_masks, attention_masks)
    ]


class Contrastive(nn.Module):
    """
    Contrastive loss. Expects as input two texts and a label of either 0 or 1. If the label == 1, then the distance between the
    two embeddings is reduced. If the label == 0, then the distance between the embeddings is increased.

    Parameters
    ----------
    model
        ColBERT model.
    score_metric
        ColBERT scoring function. Defaults to colbert_scores.
    size_average
        Average by the size of the mini-batch.

    Examples
    --------
    >>> from pylate import models, losses

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> loss = losses.Contrastive(model=model)

    >>> anchor = model.tokenize([
    ...     "fruits are healthy.",
    ... ], is_query=True)

    >>> positive = model.tokenize([
    ...     "fruits are good for health.",
    ... ], is_query=False)

    >>> negative = model.tokenize([
    ...     "fruits are bad for health.",
    ... ], is_query=False)

    >>> sentence_features = [anchor, positive, negative]

    >>> loss = loss(sentence_features=sentence_features)
    >>> assert isinstance(loss.item(), float)

    """

    def __init__(
        self,
        model: ColBERT,
        score_metric=colbert_scores,
        size_average: bool = True,
    ) -> None:
        super(Contrastive, self).__init__()
        self.score_metric = score_metric
        self.model = model
        self.size_average = size_average

    def forward(
        self,
        sentence_features: Iterable[dict[str, Tensor]],
        labels: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Compute the Constrastive loss.

        Parameters
        ----------
        sentence_features
            List of tokenized sentences. The first sentence is the anchor and the rest are the positive and negative examples.

        """
        embeddings = [
            torch.nn.functional.normalize(
                self.model(sentence_feature)["token_embeddings"], p=2, dim=-1
            )
            for sentence_feature in sentence_features
        ]
        # handle the model being wrapped in (D)DP and so require to access module first
        skiplist = (
            self.model.skiplist
            if hasattr(self.model, "skiplist")
            else self.model.module.skiplist
        )
        masks = extract_skiplist_mask(
            sentence_features=sentence_features, skiplist=skiplist
        )

        # Note: the queries mask is not used, if added, take care that the expansion tokens are not masked from scoring (because they might be masked during encoding).
        # We might not need to compute the mask for queries but I let the logic there for now
        scores = torch.cat(
            [
                self.score_metric(embeddings[0], group_embeddings, mask)
                for group_embeddings, mask in zip(embeddings[1:], masks[1:])
            ],
            dim=1,
        )

        # create corresponding labels
        # labels = torch.arange(0, rep_anchor.size(0), device=rep_anchor.device)
        labels = torch.arange(0, embeddings[0].size(0), device=embeddings[0].device)
        # compute constrastive loss using cross-entropy over the scores

        return F.cross_entropy(
            input=scores,
            target=labels,
            reduction="mean" if self.size_average else "sum",
        )



import json
import os

import torch
from safetensors.torch import load_model as load_safetensors_model
from sentence_transformers.models import Dense as DenseSentenceTransformer
from sentence_transformers.util import import_from_string
from torch import nn

__all__ = ["Dense"]


class Dense(DenseSentenceTransformer):
    """Performs linear projection on the token embeddings to a lower dimension.

    Parameters
    ----------
    in_features
        Size of the embeddings in output of the tansformer.
    out_features
        Size of the output embeddings after linear projection
    bias
        Add a bias vector
    init_weight
        Initial value for the matrix of the linear layer
    init_bias
        Initial value for the bias of the linear layer.

    Examples
    --------
    >>> from pylate import models

    >>> model = models.Dense(
    ...     in_features=768,
    ...     out_features=128,
    ... )

    >>> features = {
    ...     "token_embeddings": torch.randn(2, 768),
    ... }

    >>> projected_features = model(features)

    >>> assert projected_features["token_embeddings"].shape == (2, 128)
    >>> assert isinstance(model, DenseSentenceTransformer)

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        activation_function=nn.Identity(),
        init_weight: torch.Tensor = None,
        init_bias: torch.Tensor = None,
    ) -> None:
        super(Dense, self).__init__(
            in_features, out_features, bias, activation_function, init_weight, init_bias
        )

    def forward(self, features: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        """Performs linear projection on the token embeddings."""
        token_embeddings = features["token_embeddings"]
        projected_embeddings = self.linear(token_embeddings)
        features["token_embeddings"] = projected_embeddings
        return features

    @staticmethod
    def from_sentence_transformers(dense: DenseSentenceTransformer) -> "Dense":
        """Converts a SentenceTransformer Dense model to a Dense model.
        Our Dense model does not have the activation function.
        """
        config = dense.get_config_dict()
        config["activation_function"] = nn.Identity()
        model = Dense(**config)
        model.load_state_dict(dense.state_dict())
        return model

    @staticmethod
    def load(input_path) -> "Dense":
        """Load a Dense layer."""
        with open(os.path.join(input_path, "config.json")) as fIn:
            config = json.load(fIn)

        config["activation_function"] = import_from_string(
            config["activation_function"]
        )()

        model = Dense(**config)

        if os.path.exists(os.path.join(input_path, "model.safetensors")):
            load_safetensors_model(model, os.path.join(input_path, "model.safetensors"))
            return model

        model.load_state_dict(
            torch.load(
                os.path.join(input_path, "pytorch_model.bin"),
                map_location=torch.device("cpu"),
            )
        )
        return model



import logging
import queue

import torch
import torch.multiprocessing as mp
from transformers import is_torch_npu_available

logger = logging.getLogger(__name__)


def _start_multi_process_pool(model, target_devices: list[str] = None) -> dict:
    """
    Starts a multi-process pool to process the encoding with several independent processes
    via :meth:`SentenceTransformer.encode_multi_process <sentence_transformers.SentenceTransformer.encode_multi_process>`.

    This method is recommended if you want to encode on multiple GPUs or CPUs. It is advised
    to start only one process per GPU. This method works together with encode_multi_process
    and stop_multi_process_pool.

    Parameters
    ----------
    target_devices
        PyTorch target devices, e.g. ["cuda:0", "cuda:1", ...], ["npu:0", "npu:1", ...], or ["cpu", "cpu", "cpu", "cpu"]. If target_devices is None and CUDA/NPU
        is available, then all available CUDA/NPU devices will be used. If target_devices is None and
        CUDA/NPU is not available, then 4 CPU devices will be used.

    Returns
    -------
    A dictionary with the target processes, an input queue, and an output queue.

    """
    if target_devices is None:
        if torch.cuda.is_available():
            target_devices = [
                "cuda:{}".format(i) for i in range(torch.cuda.device_count())
            ]
        elif is_torch_npu_available():
            target_devices = [
                "npu:{}".format(i) for i in range(torch.npu.device_count())
            ]
        else:
            logger.info("CUDA/NPU is not available. Starting 4 CPU workers")
            target_devices = ["cpu"] * 4

    logger.info(
        "Start multi-process pool on devices: {}".format(
            ", ".join(map(str, target_devices))
        )
    )

    model.to("cpu")
    model.share_memory()
    ctx = mp.get_context("spawn")
    input_queue = ctx.Queue()
    output_queue = ctx.Queue()
    processes = []

    for device_id in target_devices:
        p = ctx.Process(
            target=_encode_multi_process_worker,
            args=(device_id, model, input_queue, output_queue),
            daemon=True,
        )
        p.start()
        processes.append(p)

    return {"input": input_queue, "output": output_queue, "processes": processes}


def _encode_multi_process_worker(
    target_device: str, model, input_queue: queue.Queue, results_queue: queue.Queue
) -> None:
    """
    Internal working process to encode sentences in multi-process setup
    """
    while True:
        try:
            (
                chunk_id,
                batch_size,
                sentences,
                prompt_name,
                prompt,
                precision,
                normalize_embeddings,
                padding,
                is_query,
                pool_factor,
                protected_tokens,
            ) = input_queue.get()

            embeddings = model.encode(
                sentences,
                prompt_name=prompt_name,
                prompt=prompt,
                device=target_device,
                show_progress_bar=False,
                precision=precision,
                convert_to_numpy=True,
                batch_size=batch_size,
                normalize_embeddings=normalize_embeddings,
                padding=padding,
                is_query=is_query,
                pool_factor=pool_factor,
                protected_tokens=protected_tokens,
            )

            results_queue.put([chunk_id, embeddings])
        except queue.Empty:
            break



import itertools
from typing import Callable

import torch


class ColBERTCollator:
    """Collator for ColBERT model.

    Parameters
    ----------
    tokenize_fn
        The function to tokenize the input text.
    valid_label_columns
        The name of the columns that contain the labels: scores or labels.

    Examples
    --------
    >>> from pylate import models, utils

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2", device="cpu"
    ... )

    >>> collator = utils.ColBERTCollator(
    ...     tokenize_fn=model.tokenize,
    ... )

    >>> features = [
    ...     {
    ...         "query": "fruits are healthy.",
    ...         "positive": "fruits are good for health.",
    ...         "negative": "fruits are bad for health.",
    ...         "label": [0.7, 0.3]
    ...     }
    ... ]

    >>> features = collator(features=features)

    >>> fields = [
    ...     "query_input_ids",
    ...     "positive_input_ids",
    ...     "negative_input_ids",
    ...     "query_attention_mask",
    ...     "positive_attention_mask",
    ...     "negative_attention_mask",
    ...     "query_token_type_ids",
    ...     "positive_token_type_ids",
    ...     "negative_token_type_ids",
    ... ]

    >>> for field in fields:
    ...     assert field in features
    ...     assert isinstance(features[field], torch.Tensor)
    ...     assert features[field].ndim == 2

    """

    def __init__(
        self, tokenize_fn: Callable, valid_label_columns: list[str] | None = None
    ) -> None:
        self.tokenize_fn = tokenize_fn

        if valid_label_columns is None:
            valid_label_columns = ["label", "scores"]

        self.valid_label_columns = valid_label_columns

    def __call__(self, features: list[dict]) -> dict[str, torch.Tensor]:
        """Collate a list of features into a batch."""
        batch = {"return_loss": True}

        columns = list(features[0].keys())

        if "dataset_name" in columns:
            columns.remove("dataset_name")
            batch["dataset_name"] = features[0]["dataset_name"]

        # Extract the target label.
        for label_column in self.valid_label_columns:
            if label_column in columns:
                batch["label"] = torch.tensor([row[label_column] for row in features])
                columns.remove(label_column)
                break

        # Tokenize the text.
        for column in columns:
            # We do not tokenize columns containing the ids. It would be better to throw them away during the dataset processing (TODO), but this break sentence transformers datasets extraction.
            if "_id" not in column:
                # We tokenize the query differently than the documents, TODO: define a parameter "query_column"
                is_query = "query" in column or "anchor" in column
                texts = [row[column] for row in features]
                # Flatten the list of texts if it is a list of lists (e.g, documents)
                if isinstance(texts[0], list):
                    texts = list(itertools.chain(*texts))
                tokenized = self.tokenize_fn(
                    texts,
                    is_query=is_query,
                    pad_document=True,
                )
                for key, value in tokenized.items():
                    batch[f"{column}_{key}"] = value

        return batch



from abc import ABC, abstractmethod


class Base(ABC):
    """Base class for all indexes. Indexes are used to store and retrieve embeddings."""

    @abstractmethod
    def __init__(
        self,
        name: str | None = "colbert_collection",
        recreate: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def add_documents(
        self,
        documents_ids: list[str],
        documents_embeddings: list[list[list[int | float]]],
        batch_size: int,
    ) -> None:
        pass

    @abstractmethod
    def remove_documents(self, documents_ids: list[str]) -> None:
        pass

    @abstractmethod
    def __call__(self, queries_embeddings: list[list[int | float]], k: int = 5):
        pass

    @abstractmethod
    def get_documents_embeddings(
        self, documents_ids: list[list[str]]
    ) -> list[list[list[int | float]]]:
        pass



"""Tests the training loop."""

import os
import shutil

from datasets import load_dataset
from sentence_transformers import (
    SentenceTransformerTrainer,
    SentenceTransformerTrainingArguments,
)

from pylate import losses, models, utils


def test_kd_training() -> None:
    """Test knowledge distillation training."""
    if os.path.exists(path="tests/kd"):
        shutil.rmtree("tests/kd")

    train = load_dataset(
        path="lightonai/lighton-ms-marco-mini",
        name="train",
    )

    queries = load_dataset(
        path="lightonai/lighton-ms-marco-mini",
        name="queries",
    )

    documents = load_dataset(
        path="lightonai/lighton-ms-marco-mini",
        name="documents",
    )

    train.set_transform(
        utils.KDProcessing(queries=queries, documents=documents).transform,
    )

    model = models.ColBERT(model_name_or_path="sentence-transformers/all-MiniLM-L6-v2")

    args = SentenceTransformerTrainingArguments(
        output_dir="tests/kd",
        overwrite_output_dir=True,
        num_train_epochs=1,
        per_device_train_batch_size=1,
        fp16=False,
        bf16=False,
        save_strategy="epoch",
        save_steps=1,
        save_total_limit=1,
        learning_rate=3e-6,
    )

    train_loss = losses.Distillation(model=model)

    trainer = SentenceTransformerTrainer(
        model=model,
        args=args,
        train_dataset=train,
        loss=train_loss,
        data_collator=utils.ColBERTCollator(tokenize_fn=model.tokenize),
    )

    trainer.train()

    assert os.path.isdir("tests/kd")

    if os.path.exists(path="tests/kd"):
        shutil.rmtree("tests/kd")



import ast
import logging

import datasets

logger = logging.getLogger(__name__)


class KDProcessing:
    """Dataset processing class for knowledge distillation training.

    Parameters
    ----------
    queries
        Queries dataset.
    documents
        Documents dataset.
    n_scores
        Number of scores to keep for the distillation.

    Examples
    --------
    >>> from datasets import load_dataset
    >>> from pylate import utils

    >>> train = load_dataset(
    ...    path="lightonai/lighton-ms-marco-mini",
    ...    name="train",
    ...    split="train",
    ... )

    >>> queries = load_dataset(
    ...    path="lightonai/lighton-ms-marco-mini",
    ...    name="queries",
    ...    split="train",
    ... )

    >>> documents = load_dataset(
    ...    path="lightonai/lighton-ms-marco-mini",
    ...    name="documents",
    ...    split="train",
    ... )

    >>> train.set_transform(
    ...    utils.KDProcessing(
    ...        queries=queries, documents=documents
    ...    ).transform,
    ... )

    >>> for sample in train:
    ...     assert "documents" in sample and isinstance(sample["documents"], list)
    ...     assert "query" in sample and isinstance(sample["query"], str)
    ...     assert "scores" in sample and isinstance(sample["scores"], list)

    """

    def __init__(
        self, queries: datasets.Dataset, documents: datasets.Dataset, n_ways: int = 32
    ) -> None:
        self.queries = queries["train"] if "train" in queries else queries
        self.documents = documents["train"] if "train" in documents else documents
        self.n_ways = n_ways

        self.queries_index = {
            query_id: i for i, query_id in enumerate(iterable=self.queries["query_id"])
        }

        self.documents_index = {
            document_id: i
            for i, document_id in enumerate(iterable=self.documents["document_id"])
        }

    def transform(self, examples: dict) -> dict:
        """Update the input dataset with the queries and documents."""
        if isinstance(examples["scores"][0], str):
            examples["scores"] = [
                ast.literal_eval(node_or_string=score) for score in examples["scores"]
            ]

        examples["scores"] = [score[: self.n_ways] for score in examples["scores"]]

        if isinstance(examples["document_ids"][0], str):
            examples["document_ids"] = [
                ast.literal_eval(node_or_string=document_ids)
                for document_ids in examples["document_ids"]
            ]

        examples["document_ids"] = [
            document_ids[: self.n_ways] for document_ids in examples["document_ids"]
        ]

        examples["query"] = [
            self.queries[self.queries_index[query_id]]["text"]
            for query_id in examples["query_id"]
        ]

        examples["documents"] = []
        for document_ids in examples["document_ids"]:
            documents = []
            for document_id in document_ids:
                try:
                    documents.append(
                        self.documents[self.documents_index[document_id]]["text"]
                    )
                except KeyError:
                    documents.append("")
                    logger.warning(f"Unable to find document: {document_id}")

            examples["documents"].append(documents)

        return examples

    def map(self, example: dict) -> dict:
        """Process a single example.

        Parameters
        ----------
        example
            Example to process.

        Examples
        --------
        >>> from datasets import load_dataset
        >>> from pylate import utils

        >>> train = load_dataset(
        ...    path="lightonai/lighton-ms-marco-mini",
        ...    name="train",
        ...    split="train",
        ... )

        >>> queries = load_dataset(
        ...    path="lightonai/lighton-ms-marco-mini",
        ...    name="queries",
        ...    split="train",
        ... )

        >>> documents = load_dataset(
        ...    path="lightonai/lighton-ms-marco-mini",
        ...    name="documents",
        ...    split="train",
        ... )

        >>> train = train.map(
        ...    utils.KDProcessing(
        ...        queries=queries, documents=documents
        ...    ).map,
        ... )

        >>> for sample in train:
        ...     assert "documents" in sample and isinstance(sample["documents"], list)
        ...     assert "query" in sample and isinstance(sample["query"], str)
        ...     assert "scores" in sample and isinstance(sample["scores"], list)


        """
        if isinstance(example["scores"], str):
            example["scores"] = ast.literal_eval(node_or_string=example["scores"])

        example["scores"] = example["scores"][: self.n_ways]

        if isinstance(example["document_ids"], str):
            example["document_ids"] = ast.literal_eval(
                node_or_string=example["document_ids"]
            )

        example["document_ids"] = example["document_ids"][: self.n_ways]

        processed_example = {
            "scores": example["scores"],
            "query": self.queries[self.queries_index[example["query_id"]]]["text"],
        }

        documents = []
        for document_id in example["document_ids"]:
            try:
                documents.append(
                    self.documents[self.documents_index[document_id]]["text"]
                )
            except KeyError:
                documents.append("")
                logger.warning(f"Unable to find document: {document_id}")

        processed_example["documents"] = documents

        return processed_example



from pylate import indexes, models, retrieve


def test_voyager_index(**kwargs) -> None:
    """Test the Voyager index class."""

    model = models.ColBERT(model_name_or_path="sentence-transformers/all-MiniLM-L6-v2")

    configurations = [
        {
            "sentences": ["fruits are healthy.", "fruits are good for health."],
            "convert_to_tensor": True,
            "batch_size": 1,
            "document_ids": ["1", "2"],
        },
        {
            "sentences": ["fruits are healthy.", "fruits are good for health."],
            "convert_to_tensor": False,
            "batch_size": 1,
            "document_ids": ["1", "2"],
        },
        {
            "sentences": "fruits are healthy.",
            "convert_to_tensor": True,
            "batch_size": 1,
            "document_ids": ["1"],
        },
        {
            "sentences": "fruits are healthy.",
            "convert_to_tensor": False,
            "batch_size": 1,
            "document_ids": ["1"],
        },
    ]

    for configuration in configurations:
        documents_ids = configuration.pop("document_ids")

        index = indexes.Voyager(
            index_folder="test_indexes",
            index_name="colbert",
            override=True,
            embedding_size=128,
        )

        documents_embeddings = model.encode(
            **configuration,
            is_query=False,
        )

        index.add_documents(
            documents_ids=documents_ids, documents_embeddings=documents_embeddings
        )

        queries_embeddings = model.encode(
            **configuration,
            is_query=True,
        )

        matchs = index(queries_embeddings, k=5)

        assert isinstance(matchs, dict)
        assert "documents_ids" in matchs
        assert "distances" in matchs

        assert (
            matchs["distances"].shape[0] == len(configuration["sentences"])
            if isinstance(configuration["sentences"], list)
            else 1
        )

        retriever = retrieve.ColBERT(index=index)
        results = retriever.retrieve(
            queries_embeddings=queries_embeddings,
            k=5,
        )

        assert isinstance(results, list)
        assert (
            len(results) == len(configuration["sentences"])
            if isinstance(configuration["sentences"], list)
            else 1
        )

        for query_results in results:
            for result in query_results:
                assert "id" in result
                assert "score" in result



import setuptools

from pylate.__version__ import __version__

with open(file="README.md", mode="r", encoding="utf-8") as fh:
    long_description = fh.read()

base_packages = [
    "sentence-transformers >= 3.0.1",
    "datasets >= 2.20.0",
    "accelerate >= 0.31.0",
    "voyager >= 2.0.9",
    "sqlitedict >= 2.1.0",
    "pandas >= 2.2.1",
]


dev = [
    "ruff >= 0.4.9",
    "pytest-cov >= 5.0.0",
    "pytest >= 8.2.1",
    "pandas >= 2.2.1",
    "mkdocs-material == 9.5.32",
    "mkdocs-awesome-pages-plugin == 2.9.3",
    "mkdocs-jupyter == 0.24.8",
    "mkdocs_charts_plugin == 0.0.10",
    "numpydoc == 1.8.0",
]

eval = ["ranx >= 0.3.16", "beir >= 2.0.0"]


setuptools.setup(
    name="pylate",
    version=f"{__version__}",
    license="",
    author="LightON",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lightonai/pylate",
    keywords=[],
    packages=setuptools.find_packages(),
    install_requires=base_packages,
    extras_require={
        "eval": base_packages + eval,
        "dev": base_packages + dev + eval,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



import logging

import numpy as np
import torch

from ..indexes import Voyager
from ..rank import rerank
from ..utils import iter_batch

logger = logging.getLogger(__name__)


class ColBERT:
    """ColBERT retriever.

    Parameters
    ----------
    index:
        The index to use for retrieval.

    Examples
    --------
    >>> from pylate import indexes, models, retrieve

    >>> model = models.ColBERT(
    ...     model_name_or_path="sentence-transformers/all-MiniLM-L6-v2",
    ...     device="cpu",
    ... )

    >>> documents_ids = ["1", "2"]

    >>> documents = [
    ...     "fruits are healthy.",
    ...     "fruits are good for health.",
    ... ]

    >>> documents_embeddings = model.encode(
    ...     sentences=documents,
    ...     batch_size=1,
    ...     is_query=False,
    ... )

    >>> index = indexes.Voyager(
    ...     index_folder="test_indexes",
    ...     index_name="colbert",
    ...     override=True,
    ...     embedding_size=128,
    ... )

    >>> index = index.add_documents(
    ...     documents_ids=documents_ids,
    ...     documents_embeddings=documents_embeddings,
    ... )

    >>> retriever = retrieve.ColBERT(index=index)

    >>> queries_embeddings = model.encode(
    ...     ["fruits are healthy.", "fruits are good for health."],
    ...     batch_size=1,
    ...     is_query=True,
    ... )

    >>> results = retriever.retrieve(
    ...     queries_embeddings=queries_embeddings,
    ...     k=2,
    ...     device="cpu",
    ... )

    >>> assert isinstance(results, list)
    >>> assert len(results) == 2

    >>> queries_embeddings = model.encode(
    ...     "fruits are healthy.",
    ...     batch_size=1,
    ...     is_query=True,
    ... )

    >>> results = retriever.retrieve(
    ...     queries_embeddings=queries_embeddings,
    ...     k=2,
    ...     device="cpu",
    ... )

    >>> assert isinstance(results, list)
    >>> assert len(results) == 1

    """

    def __init__(self, index: Voyager) -> None:
        self.index = index

    def retrieve(
        self,
        queries_embeddings: list[list | np.ndarray | torch.Tensor],
        k: int = 10,
        k_token: int = 100,
        device: str | None = None,
        batch_size: int = 50,
    ) -> list[list[dict]]:
        """Retrieve documents for a list of queries.

        Parameters
        ----------
        queries_embeddings
            The queries embeddings.
        k
            The number of documents to retrieve.
        k_token
            The number of documents to retrieve from the index. Defaults to `k`.
        device
            The device to use for the embeddings. Defaults to queries_embeddings device.

        """
        if k > k_token:
            logger.warning(
                f"k ({k}) is greater than k_token ({k_token}), setting k_token to k."
            )
            k_token = k
        reranking_results = []
        for queries_embeddings_batch in iter_batch(
            queries_embeddings,
            batch_size=batch_size,
            desc=f"Retrieving documents (bs={batch_size})",
        ):
            retrieved_elements = self.index(
                queries_embeddings=queries_embeddings_batch,
                k=k_token,
            )

            documents_ids = [
                list(
                    set(
                        [
                            document_id
                            for query_token_document_ids in query_documents_ids
                            for document_id in query_token_document_ids
                        ]
                    )
                )
                for query_documents_ids in retrieved_elements["documents_ids"]
            ]

            documents_embeddings = self.index.get_documents_embeddings(documents_ids)

            reranking_results.extend(
                rerank(
                    documents_ids=documents_ids,
                    queries_embeddings=queries_embeddings_batch,
                    documents_embeddings=documents_embeddings,
                    device=device,
                )
            )
        return [query_results[:k] for query_results in reranking_results]



def load_custom_dataset(path: str, split: str = "test") -> tuple[list, list, dict]:
    """Load a custom dataset.

    Parameters
    ----------
    path
        Path of the dataset.
    split
        Split to load.

    Examples
    --------
    """
    from beir.datasets.data_loader import GenericDataLoader

    documents, queries, qrels = GenericDataLoader(path).load(split=split)

    documents = [
        {
            "id": document_id,
            "text": f"{document['title']} {document['text']}".strip()
            if "title" in document
            else document["text"].strip(),
        }
        for document_id, document in documents.items()
    ]

    qrels = {
        queries[query_id]: query_documents
        for query_id, query_documents in qrels.items()
    }

    return documents, list(qrels.keys()), qrels
